var utils = require ( './utils' );
var R = require ( 'ramda' );

var animationPeriodMinutes = 5;

exports.assemblerType = 'rbb-average-touch';

exports.particleType = 'rbb-average-touch';

exports.initial = function ( config ) {
    var findTeam = utils.chooseTeamName ( R.values ( config.teams ) );
    
    var initSide = function ( side ) {
        return {
            title: findTeam ( side ),
            players: [],
            passes: []
        };
    };

    return {
        particle: {
            title: 'Average touch positions (0 min)',
            type: exports.particleType,
            assemblerType: exports.assemblerType,
            version: config.version,
            data: {
                home: initSide ( 'home' ),
                away: initSide ( 'away' )
            },
            formattedEndTime: '0 min',
            endTime: 0
        },
        playerCounts: {
            home: {},
            away: {}
        },
        passCounts: {
            home: {},
            away: {}
        },
        teams: config.teams,
        gameTime: null,
        prevEvent: null,
        prevEventIndex: null
    };
};

exports.reducer = function ( context, event ) {
    if ( event.type.name === 'End' ) {
        return R.evolve ( {
            particle: {
                title: function ( title ) {
                    return {
                        1: 'Average touch positions (half time)',
                        2: 'Average touch positions (full time)',
                        3: 'Average touch positions (half time in extra time)',
                        4: 'Average touch positions (end of extra time)'
                    }[ event.period_id ] || 'Average touch positions (' + utils.getFormattedTime ( event ) + ')';
                }
            }
        }, context );
    }

    utils.extendTeamData ( context, event );
    utils.extendPlayerData ( context, event );

    if ( ! utils.isTouchEvent ( event ) ) {
        return context;
    }

    if ( ! ( event.team && event.player && event.x && event.y ) ) {
        return context;
    }

    var team = context.teams[event.team.id];
    var side = team.side.toLowerCase();
    var eventIndex = Math.floor ( event.min / animationPeriodMinutes );

    var transformPlayerCounts = R.curry ( function ( event, playerCounts ) {
        var playerCount = playerCounts[event.player.id];

        return R.merge ( playerCounts, R.fromPairs ( R.of ( R.pair ( event.player.id, {
            mins: R.concat ( playerCount ? playerCount.mins : [], [ event.min ] ),
            count: ( playerCount ? playerCount.count : 0 ) + 1,
            X: ( playerCount ? playerCount.X : 0 ) + parseFloat ( event.x ),
            Y: ( playerCount ? playerCount.Y : 0 ) + parseFloat ( event.y )
        } ) ) ) );
    } );

    var transformPassCounts = R.curry ( function ( prevEvent, event, passCounts ) {
        if ( ! prevEvent || prevEvent.team.id !== event.team.id || prevEvent.player.id === event.player.id ) {
            return passCounts;
        }

        var passId = [ prevEvent.player.id, event.player.id ].sort ().join ( '-' );
        var passCount = passCounts[passId];

        return R.merge ( passCounts, R.fromPairs ( R.of ( R.pair ( passId, {
            mins: R.concat ( passCount ? passCount.mins : [], [ prevEvent.min ] ),
            count: ( passCount ? passCount.count : 0 ) + 1
        } ) ) ) );
    } );

    var tX = R.partial ( utils.translateX, [ 'Left to Right' ] );
    var tY = R.partial ( utils.translateY, [ 'Left to Right' ] );

    var getPlayerXY = function ( playerCount ) {
        return {
            x: tX ( playerCount.X / playerCount.count ),
            y: tY ( playerCount.Y / playerCount.count )
        };
    };

    var getPlayer = function ( players, playerId, display ) {
        var player = players[playerId];

        return {
            id: playerId,
            name: player ? player.name : '',
            jersey: player ? player.jersey_number : '',
            substitute: player ? player.substitute : false,
            display: display
        };
    };

    var transformPlayers = function ( eventIndex, prevEventIndex, side, players, playerCounts, oldPlayers ) {
        return R.map ( function ( playerPair ) {
            var position = getPlayerXY ( playerPair[1] );

            var oldPlayer = R.find ( function ( player ) {
                return player.player.id === playerPair[0];
            }, oldPlayers );

            var anim = [];

            if ( oldPlayer && oldPlayer.anim && oldPlayer.anim.length ) {
                anim = R.clone ( oldPlayer.anim );
            } else {
                anim = R.times ( R.always ( {
                    start: { x: tX ( 50 ), y: tY ( 50 ) },
                    end: { x: tX ( 50 ), y: tY ( 50 ) },
                    weight: 0
                } ), eventIndex );
            }

            while ( anim.length < eventIndex ) {
                anim.push ( R.clone ( R.last ( anim ) ) );
            }

            anim[eventIndex] = {
                start: position,
                end: position,
                weight: 0.5 + playerPair[1].count * 0.015,
            };

            return {
                type: 'dribble',
                side: side,
                player: getPlayer ( players, playerPair[0], true ),
                start: position,
                end: position,
                weight: 0.5 + playerPair[1].count * 0.015,
                anim: anim
            };
        }, R.toPairs ( playerCounts ) );
    };

    var transformPasses = function ( eventIndex, prevEventIndex, side, players, playerCounts, passCounts, oldPasses ) {
        return R.map ( function ( passPair ) {
            var playerId = R.head ( R.split ( '-', passPair[0] ) ), playerCount = playerCounts[playerId];
            var playerToId = R.last ( R.split ( '-', passPair[0] ) ), playerToCount = playerCounts[playerToId];
            var start = getPlayerXY ( playerCount ), end = getPlayerXY ( playerToCount );

            var oldPass = R.find ( function ( pass ) {
                return pass.player.id === playerId && pass.playerTo.id === playerToId;
            }, oldPasses );

            var anim = [];

            if ( oldPass && oldPass.anim && oldPass.anim.length ) {
                anim = R.clone ( oldPass.anim );
            } else {
                anim = R.times ( R.always ( {
                    start: { x: tX ( 50 ), y: tY ( 50 ) },
                    end: { x: tX ( 50 ), y: tY ( 50 ) },
                    weight: 0
                } ), eventIndex );
            }

            while ( anim.length < eventIndex ) {
                anim.push ( R.clone ( R.last ( anim ) ) );
            }

            anim[eventIndex] = {
                start: start,
                end: end,
                weight: passPair[1].count,
            };

            return {
                type: 'vertex',
                side: side,
                player: getPlayer ( players, playerId, false ),
                playerTo: getPlayer ( players, playerToId, false ),
                start: start,
                end: end,
                weight: passPair[1].count,
                anim: anim
            };
        }, R.toPairs ( passCounts ) );
    };

    var playerCounts = transformPlayerCounts ( event, context.playerCounts[side] );

    var passCounts = transformPassCounts ( context.prevEvent, event, context.passCounts[side] );

    var players = transformPlayers ( eventIndex, context.prevEventIndex, side, context.players, playerCounts, context.particle.data[side].players );

    var passes = transformPasses ( eventIndex, context.prevEventIndex, side, context.players, playerCounts, passCounts, context.particle.data[side].passes );

    return R.evolve ( {
        particle: {
            title: R.always ( 'Average touch positions (' + utils.getFormattedTime ( event ) + ')' ),
            data: R.fromPairs ( R.of ( R.pair ( side, {
                players: R.always ( players ),
                passes: R.always ( passes )
            } ) ) ),
            formattedEndTime: function ( oldTime ) {
                if ( ( event.min * 60 + event.sec ) > context.particle.endTime ) {
                    return utils.getFormattedTime ( event );
                }

                return oldTime;
            },
            endTime: function ( time ) {
                var newTime = event.min * 60 + event.sec;

                return Math.max ( time, newTime );
            }
        },
        playerCounts: R.fromPairs ( R.of ( R.pair ( side, R.always ( playerCounts ) ) ) ),
        passCounts: R.fromPairs ( R.of ( R.pair ( side, R.always ( passCounts ) ) ) ),
        prevEvent: R.always ( event ),
        prevEventIndex: R.always ( eventIndex )
    }, context );
};
