var utils = require('./utils');

var fov = 435;
var pitchWidth = (105 * 2);
var pitchHeight = (68 * 2) - 6;

var halfWidth = pitchWidth/2;
var halfHeight = pitchHeight/2;

var angleInDegrees = 75;
var numKicks = 10;
var kickHeight = 60;

var angle = degreesToRadians(angleInDegrees);

var cos = Math.cos(angle);
var sin = Math.sin(angle);

var rotZ = degreesToRadians(-90);
var sinRotZ = Math.sin(rotZ);
var cosRotZ = Math.cos(rotZ);

var assemblerType = 'rbb-corners-and-free-kicks';

exports.particleType = 'rbb-global-thermo-half-pitch';

exports.initial = function (config) {
    return {
        particle: {
            title: 'Corners and Free Kicks',
            type: exports.particleType,
            assemblerType: assemblerType,
            version: config.version,
            uniqueId: [ assemblerType, 0 ].join ( '-' ),
            data: {
                events: []
            }
        },
        interrupted: false,
        shots: [],
        kicks: [],
        meta: {
            publishTags: ['reducer:rbb-corners-and-free-kicks'],
            generatedTimeStamp: null,
            eventTimeStamp: null
        },
	side: config.side,
        teams: config.teams,
        players: {},
        gameTime: null
    };
};

exports.reducer = function (context, event) {
    utils.extendTeamData(context, event);
    utils.extendPlayerData(context, event);
    if (event.team) {
	var team = context.teams[event.team.id];
	var side = team.side.toLowerCase();
    }
    
    if (utils.isCornerOrFreeKick(event) && (side === context.side) ) {
        var team_name = context.teams[event.team.id].name;
        context.particle.title = team_name + ' corners and free kicks';
        var end_x = event.qualifiers['pass-end-x'].value;
        var end_y = event.qualifiers['pass-end-y'].value;
        var player = context.players[event.player.id];
        var direction = 'Left to Right';

        var abs_start_x = utils.absoluteX(direction, event.x);
        var abs_start_y = utils.absoluteY(direction, event.y);
        var abs_end_x = utils.absoluteX(direction, end_x);
        var abs_end_y = utils.absoluteY(direction, end_y);

        // only interested in opposition half
        if(abs_start_x > 55 && abs_end_x > 55) {
            var qualifiers = utils.getEventQualifiersAsBooleans(event);    

            context.kicks.push({
                type: 'pass',
                player: player ? player.jersey_number : '',
                start: {
                    x: utils.translateX(direction, event.x),
                    y: utils.translateY(direction, event.y)
                },
                end: {
                    x: utils.translateX(direction, end_x),
                    y: utils.translateY(direction, end_y)
                },
                showCircle: true
            });

            // going through in steps to convert 2d point to rotated 3d point displayed in our world
            var shot = [];

            // adjust the raw coords to midfield being 0,0
            shot[0] = [
                convertToMidfieldOrigin(abs_start_x, pitchWidth, 'x') - 12,
                convertToMidfieldOrigin(abs_start_y, pitchHeight, 'y'),
                0
            ];
            shot[1] = [
                convertToMidfieldOrigin(abs_end_x, pitchWidth, 'x') - 12,
                convertToMidfieldOrigin(abs_end_y, pitchHeight, 'y'),
                0
            ];

            // add the midpoint which will have a z
            shot[2] = calculateMidpoint( shot[0], shot[1], qualifiers.chipped );

            // convert the points to rotated 3d points
            var shotConvertedTo3D = [
                get3DPosition(shot[0], sin, cos),
                get3DPosition(shot[1], sin, cos),
                get3DPosition(shot[2], sin, cos)
            ];

            var type = (qualifiers['corner-taken']) ? 'corner' : 'pass';
            // convert the 3d points to our world coords
            // cp = control point... derived from the midpoint
            context.shots.push({
                type: type,
                qualifiers: qualifiers,
                start: convertPoint( shotConvertedTo3D[0] ),
                end:  convertPoint( shotConvertedTo3D[1] ),
                cp:  convertPoint( shotConvertedTo3D[2] )
            });
            context.particle.data.events = context.shots;
            context.last_team = event.team.id;
            context.particle.uniqueId = [ assemblerType, event.id ].join ( '-' );
        }
    }
    else if (utils.isPassChainEnd(event, context.last_team)) {
        context.interrupted = true;
    }
    context.gameTime = utils.createGameTimeObject(event, context.gameTime);
    return context;
};

function degreesToRadians(degrees) {
    return degrees * Math.PI / 180;
}

function convertToMidfieldOrigin(value, size, axis) {
    return (Number(value) / 100 * size) - (size/2);
}

function calculateMidpoint( startPt, endPt, chipped ) {
    // startPt and endPt are arrays representing field position
    // [x,y,z]. z will be 0 adjust height based on distance roughly
    var adjustedKickHeight = (chipped) ? kickHeight : 18;
    var dist = parseInt(Math.max( Math.abs(startPt[0] - endPt[0]), Math.abs(startPt[1] - endPt[1]) ));
    var adjustedHeight = utils.clamp( Math.round(mapNumberRange(dist, 0, pitchWidth, 0, adjustedKickHeight)), 0, 24);
    return [ (startPt[0] + endPt[0]) / 2, (startPt[1] + endPt[1]) / 2, -adjustedHeight ];
}

function convertPoint(pt) {
    var x3d = pt[0];
    var y3d = pt[1];
    var z3d = pt[2];
    var scale = fov/(z3d + fov);
    var x2d = (x3d * scale) + halfWidth; 
    var y2d = (y3d * scale) + halfHeight;
    return { x: x2d, y: y2d };
}

function get3DPosition(loc, sin, cos) {
    var newLoc = rotateNegativeZAxis( loc, sinRotZ, cosRotZ );
    newLoc = shiftYAxis( newLoc );
    var pos = [];
    pos[0] = 1 * newLoc[0];
    pos[1] = (cos * newLoc[1]) + (sin * newLoc[2]);
    pos[2] = (-sin * newLoc[1]) + (cos * newLoc[2]);
    return pos;
}

function rotateNegativeZAxis(loc, sin, cos) {
    // cos θ  −sin θ   0
    // sin θ   cos θ   0
    // 0    0   1
    var x = loc[0];
    var y = loc[1];
    var z = loc[2];
    var x1 = Math.round( (x * cos) + (-sin * y) + (0 * z));
    var y1 = Math.round( (sin * x) + (cos * y) + (0 * z) );
    var z1 = (0 * x) + (0 * y) + (1 * z);
    return [x1, y1, z1 ];
}

function shiftYAxis(loc) {
    return [loc[0], loc[1] + halfWidth + 200, loc[2]];
}

function mapNumberRange ( num, in_min , in_max , out_min , out_max ) {
    return ( num - in_min ) * ( out_max - out_min ) / ( in_max - in_min ) + out_min;
}

function addFakeEvents() {
    var events = [
        {
            start: {
                x: -19.32546865125147, 
                y: 165.64738986373388
            },
            cp: {
                x: 60.83651106690738,
                y: 123.93020414203176 
            },
            end: {
                x: 144.49705743249552,
                y: 172.29701842508075
            }
        },
        {
            start: {
                x: -19.32546865125147, 
                y: 165.64738986373388
            },
            cp: {
                x: 8.323045169665704,
                y: 159.82845737457143 
            },
            end: {
                x: 40.35254931308282,
                y: 191.5614080120176
            }
        },
        {
            start: {
                x: 229.32546865125147, 
                y: 165.64738986373388
            },
            cp: {
                x: 171.3833631285637,
                y: 136.7404083092537 
            },
            end: {
                x: 105,
                y: 177.3176284839344
            }
        }
    ];
    return events;
}
