var utils = require('./utils'),
    R = require('ramda');


// events that get sent twice at the same time
var ONE_OUTCOME_EVENTS = [  
    'Corner Awarded', 
    'Out', 
    'Aerial', 
    'Foul',
    'Take On'
];

// events that get sent once for each team, but we only need it once... ie Start of a period. 
var EACH_TEAM_EVENTS = [
    'Start', 
    'End', 
    'Collection End', 
    'Start delay', 
    'End delay', 
    'Team set up'
];

// event types to just ignore for dashboard timeline purposes
var IGNORE_EVENTS = [   
    'Deleted event', 
    'Offside provoked', 
    'Formation change', 
    'Injury Time Announcement', 
    'Dispossessed', 
    'Player Off', 
    'Player on', 
    'Corner Awarded',
    'Out',
    'Contentious referee decision',
    'Referee Drop Ball',
    'Coach Setup',
    'Caught Offside',
    'Chance missed',
    'Formation change',
    'Condition change'
];

// marker events for our timeline
var MARKER_EVENTS = [
    'Goal', 
    'Miss', 
    'Attempt Saved', 
    'Post',
    'Card'
];

// TODO: where is the best place to put this?
// Utility functions that only need setting up once
var isInList = R.curry( function(list, event) {
    return R.contains( R.path(['type', 'name'])(event), list);
} );

var isEachTeamEvent = isInList( EACH_TEAM_EVENTS );
var isOneOutcomeEvent = isInList( ONE_OUTCOME_EVENTS );
exports.isMarkerEvent = isInList( MARKER_EVENTS );
var isIgnoreEvent = isInList( IGNORE_EVENTS );

var isHome = R.propEq('side', 'home');
var isWinnerOutcome = R.propEq('outcome', 1);
exports.isGoal = R.pathEq(['type', 'name'], 'Goal');
var isOwnGoal = R.path(['qualifiers', 'own-goal']);
exports.isCard = R.pathEq(['type', 'name'], 'Card');

// filter out the double events to just one. use home
var onlyOneTeamFilter = R.allPass( [ isEachTeamEvent, isHome ] );

// only interested in the 'winner' event of the oneOutcomeEvents. filter out 'loser' ones
var onlyWinnerFilter = R.either( R.compose( R.not, isOneOutcomeEvent ), isWinnerOutcome );

var goalLineValue = R.ifElse(
    R.pathOr(null, ['qualifiers','goal-mouth-y-co-ordinate']),
    R.always( 100 ),
    R.always( 10 )
);


var getHighest = function(nums) {
  return R.reduce( R.max, 0, R.map( parseInt, nums ) );
};

// is the event useful to plot in the timeline?
exports.isTimelineEvent = R.allPass([
    R.compose( R.not, isIgnoreEvent ),
    R.compose( R.not, isEachTeamEvent ),
    onlyWinnerFilter
]);

exports.getCardType = R.cond([
  [R.path(['qualifiers', 'second-yellow']),   R.always('second-yellow')],
  [R.path(['qualifiers', 'red-card']), R.always('red')],
  [R.path(['qualifiers', 'yellow-card']), R.always('yellow')],
  [R.T, R.always('') ]
]);

var combineKeyedArrays = R.pipe(
    R.values,
    R.unnest,
    R.filter( Boolean )
);

var convertEventForDebug = function(event, idx) {
    var obj = R.pick( ['min', 'sec', 'type', 'qualifiers'])(event);
    return R.pick( ['min', 'sec', 'type', 'qualifiers'])(event);
    // obj.qualifiers = R.map( R.ifElse(R.isNil, R.always(''), R.identity), R.pluck('value')( event.qualifiers ) );
    // return obj;
};

var getInjuryTimeMinutes = R.ifElse(
    R.pathEq(['type', 'name'], 'Injury Time Announcement'),
    R.compose( parseInt, R.max(0), R.dec, R.path(['qualifiers', 'minutes', 'value'])),
    R.always( 0 )
);

exports.isGamePeriodEvent = R.compose( R.allPass( [ R.lt( R.__, 6 ), R.gt( R.__, 0) ] ), R.prop('period_id') );
var isNotGamePeriodEvent = R.compose( R.not, exports.isGamePeriodEvent );

var slugify = R.compose( R.join('-'), R.split(' '), R.toLower); 
exports.slugifyEvent = R.compose( slugify, R.path(['type', 'name']) );

var getDepth = function(event) {
    return getHighest(
      [ R.prop('x', event), 
        R.pathOr(10, ['qualifiers', 'pass-end-x', 'value'], event),
        goalLineValue(event)
      ]);    
};

var getTop = function(num) {
    return (100 - num) / 2;
};

// END UTILITY FUNCTIONS


var assemblerType = 'rbb-dashboard';

exports.particleType = 'rbb-dashboard-timeline';

exports.initial = function (config) {

    var findTeam = utils.chooseTeamName (R.values(config.teams));

    return {
        particle: {
            title: 'Match DNA',
            type: exports.particleType,
            assemblerType: assemblerType,
            version: config.version,
            uniqueId: [ assemblerType, 0 ].join ( '-' ),
            data: {
                "match-time": "",
                home: {
                    name: findTeam('home'),
                    score: 0,
                    penaltyScore: 0
                },
                away: {
                    name: findTeam('away'),
                    score: 0,
                    penaltyScore: 0
                },
                periods: {},
                hasExtraPlay: false,
                hasPenalties: false,               
            }
        },
        eventIndex: -1,
        longestPeriod: 45,              
        teams: config.teams,
        debug: config.debug,
        allEvents: []
    };
};


exports.normalizeMinutes = function(event) {
    
    var expectedMinOffset = {
        "1": 0,
        "2": 45,
        "3": 90,
        "4": 105,
        "5": 120
    };

    var eventMin = ( event.min - expectedMinOffset[event.period_id] ) + (event.sec/60);
    var injuryTime = getInjuryTimeMinutes(event); 
    return eventMin + injuryTime;
};

exports.reducer = function (context, event) {
    // keep a count of the sequence so we can skip to a specific event from the UI
    context.eventIndex = R.add(1, R.prop('eventIndex', context ) );
    
    var period = R.prop('period_id', event);
    var periodLength = R.max( (period < 3) ? context.longestPeriod : (period < 5) ? 15 : 5, exports.normalizeMinutes(event) );

    // store latest time
    // ignoring all non game period events for dashboard time
    if( exports.isGamePeriodEvent( event ) ) {

        if( period > 4 ) {
                context.particle.data.hasPenalties = true;
        } else {
            context.particle.data['match-time'] = utils.getFormattedTime(event);   

            if( period > 2 ) {
                context.particle.data.hasExtraPlay = true;                 
            }
        }        
    }    
    

    var side = context.teams[event.team.id].side.toLowerCase();
    var otherSide = (side === 'home') ? 'away' : 'home';

    // handle the events that belong in the timeline
    if( exports.isTimelineEvent(event) ) {
        var type = exports.slugifyEvent( event ) ;
        if( type === 'goal' ) {
            var scoreType = (event.period_id === 5) ? 'penaltyScore' : 'score';

            if (event.qualifiers['own-goal']) {
                context.particle.data[otherSide][scoreType] ++;                                                
            } else {
                context.particle.data[side][scoreType] ++;                                
            }
        }

        var depth = getDepth( event );

        var timelineEvent = {
            side: side,
            type: type,
            hint: R.path('type', 'name'),
            idx: R.prop('eventIndex', context ),
            is: {
                home: R.equals( side, 'home' ),
                goal: exports.isGoal(event),
                ownGoal: R.has('own-goal', event.qualifiers),
                card: exports.isCard(event),
                marker: exports.isMarkerEvent(event),
                display: exports.isTimelineEvent(event)
            },
            min: event.min,
            sec: event.sec,
            period_id: event.period_id,
            card: exports.getCardType(event),
            timeline: {
                depth: depth,
                top: getTop(depth),
                pct: {
                    x: ( exports.normalizeMinutes(event) / periodLength) * 100
                }
            }
        };
        var bucket = context.particle.data.periods['p' + event.period_id] || [];
        context.particle.data.periods['p' + event.period_id] = R.concat( bucket, [timelineEvent] );
        context.particle.uniqueId = [ assemblerType, event.id ].join ( '-' );
    }


    // adjust all the timeline positions, either when there is an injury time announcement or if an event creeps beyond the latest known time relative to a 45 minute half
    if(periodLength > context.longestPeriod) {
       R.forEach( function( item ) {
            item.timeline.pct.x = ( exports.normalizeMinutes(item) / periodLength) * 100;
        }, context.particle.data.periods['p' + event.period_id]);
        context.longestPeriod = periodLength;
    }

    if( context.debug ) {
        context.allEvents.push( 
            R.merge(
                R.pick(['x', 'y', 'min', 'sec', 'period_id', 'player', 'outcome'])(event),
                {     
                    idx: context.eventIndex,
                    name: event.type.name,       
                    qualifiers: R.map( R.ifElse(R.isNil, R.always(''), R.identity), R.pluck('value')( event.qualifiers ) )
                }
            )
        );
    }
    else {
        context.allEvents.push(
            {
                idx: context.eventIndex,
                name: event.type.name
            }
        );
    }


    return context;
};
