var utils = require('./utils');
var R = require('ramda');

var assemblerType = 'rbb-event-chains-on-pitch';

exports.particleType = 'rbb-single-pitch-map';

function resetPassChain(context) {
    context.interrupted = false;
    context.particle.data.events = [];
    context.particle.data.animated = false;
    context.particle.title = 'Passes';
    context.particle.data.goalkeepers = [];
}

// test whether two events should be joined by a dribble line
function joinWithDribble(a, b) {
    if (!a.team || !b.team || a.team.id !== b.team.id) {
	return false;
    }
    if (a.type.name === 'Pass') {
	// pass -> pass
	// pass -> dispossesed
	// pass -> post
	// pass -> miss
	// pass -> goal
	// pass -> error
	// pass -> take on
        return R.contains(b.type.name, [
	    'Pass',
	    'Dispossessed',
	    'Post',
	    'Miss',
	    'Goal',
	    'Error',
	    'Take On'
	]);
    }
    // interception -> pass
    else if (a.type.name === 'Interception' && a.outcome === 1) {
	return b.type.name === 'Pass';
    }
    // tackle -> pass
    else if (a.type.name === 'Tackle' && a.outcome === 1) {
	return b.type.name === 'Pass';
    }
    return false;
}

function calculateDribble(context, event) {
    var events = context.particle.data.events;
    if (!context.last_plotted_event || events.length === 0) {
	return false;
    }
    if (joinWithDribble(context.last_plotted_event, event)) {
	var prev_event = events[events.length - 1];
	var last_x = prev_event.start.x;
	var last_y = prev_event.start.y;
	if (prev_event.end) {
	    last_x = prev_event.end.x;
	    last_y = prev_event.end.y;
	}
	var team = context.teams[event.team.id];
    var player = event.player && context.players[event.player.id];
        return {
            type: 'dribble',
            side: team.side.toLowerCase(),
            qualifiers: {},
            player: {
		        name: player ? player.name : '',
                jersey: player ? player.jersey_number : '',
                display: true
            },
            start: {
                x: last_x,
                y: last_y
            },
            end: {
                x: utils.translateX(team.direction, event.x),
                y: utils.translateY(team.direction, event.y)
            }
        };
    }
    return false;
}

function addToChain(context, data) {
    var opposing_side = data.side !== context.currentSide;
    if (context.interrupted === 'always') {
	resetPassChain(context);
    }
    else if (context.interrupted === 'maybe') {
	var len = context.particle.data.events.length - 1;
	var last_side = context.currentSide;
	if (len > 0) {
	    var last_event = context.particle.data.events[len - 1];
	    last_side = last_event.side.toLowerCase();
	}
	if (opposing_side && last_side !== context.currentSide) {
	    // confirmed, it was a real interruption
	    resetPassChain(context);
	}
	else {
	    // keep the chain going
	    context.interrupted = false;
	}
    }
    else if (opposing_side) {
	// swapped sides without knowing why, reset the chain just to be safe
	resetPassChain(context);
    }
    if (context.particle.data.events.length === 0) {
	// first event in chain
	context.currentSide = data.side;
    }
    context.particle.data.events.push(data);
    context.particle.data.animated = context.particle.data.events.length > 3;
}

function extractQualifierFlags(event) {
    var flags = {};
    for (var k in event.qualifiers) {
        var q = event.qualifiers[k];
        if (q.type.values === '') {
            flags[k] = true;
        }
    }
    return flags;
}

function extendChain(context, event) {
    if (R.contains(event.type.name, ['Goal', 'Post', 'Miss', 'Attempt Saved'])) {
    var player = event.player && context.players[event.player.id];
	var side = context.teams[event.team.id].side.toLowerCase();
        var otherSide = utils.getOpposingSide(side);
        var time = utils.getFormattedTime(event);

        var team = context.teams[event.team.id];
        context.particle.data.goalkeepers = [];

        context.eventTime = utils.createGameTimeObject( event );

        var jersey =  player ? player.jersey_number : '';
        var direction = context.teams[event.team.id].direction;

        context.particle.data['attack-left'] = (direction === 'Right to Left');
        context.particle.data['action-left'] = (direction === 'Right to Left');
        if (event.qualifiers['own-goal']) {
            // action is happening at opposite goal line
            context.particle.data['action-left'] = !(context.particle.data['action-left']);
        }

	var dribble = calculateDribble(context, event);
	if (dribble) {
	    addToChain(context, dribble);
	}

	var start = {
            "x": utils.translateX(direction, event.x),
            "y": utils.translateY(direction, event.y)
	};
	var end;
	if (event.qualifiers.hasOwnProperty('blocked-x-co-ordinate')) {
	    end = {
		"x": utils.translateX(
		    direction,
		    event.qualifiers['blocked-x-co-ordinate'].value
		),
		"y": utils.translateY(
		    direction,
		    event.qualifiers['blocked-y-co-ordinate'].value
		)
	    };
	}
	else {
	    end = {
		"x": utils.translateX(direction, event.qualifiers['own-goal'] ? 0 : 100),
		"y": utils.translateY(direction, event.qualifiers['goal-mouth-y-co-ordinate'].value)
	    };
	}

	// it's not clear whether a 'Save' is by the GK, plot a
	// generic marker for opposing team
	if (event.type.name === 'Attempt Saved' &&
	    event.qualifiers['blocked-x-co-ordinate'] &&
	    event.qualifiers['blocked-y-co-ordinate']) {
	    context.last_plotted_event = event;
	    addToChain(context, {
                "type": "position",
                "static": true,
                "side": utils.getOpposingSide(side),
                "start": {
                    "x": utils.translateX(
			direction,
			event.qualifiers['blocked-x-co-ordinate'].value
		    ),
                    "y": utils.translateY(
			direction,
			event.qualifiers['blocked-y-co-ordinate'].value
		    )
                },
                "player": {
                    "display": true
                }
            });
	}

	var type = 'pass';
	if (event.type.name === 'Goal') {
	    type = 'goal';
	}
	else if (event.type.name === 'Attempt Saved') {
	    type = 'shot-on-target';
        }
	else if (event.type.name === 'Miss') {
	    type = 'shot-off-target';
	}
	context.last_plotted_event = event;
        addToChain(context, {
            type: type,
            side: side,
            "own-goal": (event.qualifiers['own-goal']) ? true : false,
            "start": start,
            "end": end,
            "player": {
                name: player ? player.name : '',
                "jersey": jersey,
                "display": !dribble
            },
            "is-in-attacking-half": utils.isInAttackingHalf(event) 
        });
	
        // in an own-goal it's not clear which GK we have co-ordinates
        // for, or how they should be translated, so leave the GK off
        // the pitch map. Also, we don't get gk-x-coordinate and
        // gk-y-coordinate for 'Miss' or 'Attempt Saved' events.
        if (event.type.name === 'Goal' && !event.qualifiers['own-goal']) {
            context.particle.data.goalkeepers.push({
                "type": "position",
                "static": true,
                "side": utils.getOpposingSide(side),
                "start": {
                    "x": utils.translateX('Left to Right', event.qualifiers['gk-x-coordinate'].value),
                    "y": utils.translateY('Left to Right', event.qualifiers['gk-y-coordinate'].value)
                },
                "player": {
                    name: player ? player.name : '',
                    "jersey": "GK",
                    "display": true
                }
            });
        }
    }
    // genericPassEvent includes anything eith end x/y co-ords (e.g. Clearance)
    else if (utils.isGenericPassEvent(event)) {
	direction = context.teams[event.team.id].direction;
	context.particle.data['attack-left'] = (direction === 'Right to Left');
        context.particle.data['action-left'] = (direction === 'Right to Left');
	
	dribble = calculateDribble(context, event);
	if (dribble) {
	    addToChain(context, dribble);
	}
	
	player = event.player && context.players[event.player.id];
	var end_x = event.qualifiers['pass-end-x'].value;
        var end_y = event.qualifiers['pass-end-y'].value;
	context.last_plotted_event = event;
	addToChain(context, {
            type: event.qualifiers['throw-in'] ? 'throwin' : 'pass',
            side: context.teams[event.team.id].side.toLowerCase(),
            qualifiers: extractQualifierFlags(event),
            player: {
		        name: player ? player.name : '',
                jersey: player ? player.jersey_number : '',
                display: !dribble
            },
            start: {
                x: utils.translateX(direction, event.x),
                y: utils.translateY(direction, event.y)
            },
            end: {
                x: utils.translateX(direction, end_x),
                y: utils.translateY(direction, end_y)
            }
        });
        context.particle.uniqueId = [ assemblerType, event.id ].join ( '-' );
    }
    else if (R.contains(event.type.name, ['Interception', 'Blocked Pass', 'Aerial', 'Tackle', 'Take On', 'Dispossessed', 'Ball recovery', 'Ball touch']) || (event.type.name === 'Foul' && event.outcome === 0)) {
	player = event.player && context.players[event.player.id];
	direction = context.teams[event.team.id].direction;
	context.last_plotted_event = event;
	addToChain(context, {
            type: 'position',
            "static": true,
            side: context.teams[event.team.id].side.toLowerCase(),
            qualifiers: extractQualifierFlags(event),
            player: {
		        name: player ? player.name : '',
                jersey: player ? player.jersey_number : '',
                display: true
            },
            start: {
                x: utils.translateX(direction, event.x),
                y: utils.translateY(direction, event.y)
            }
        });
    }
    else if (event.type.name === 'Out' && event.outcome === 0) {
	player = event.player && context.players[event.player.id];
	direction = context.teams[event.team.id].direction;
	context.last_plotted_event = event;
	addToChain(context, {
            type: 'out',
            "static": true,
            side: context.teams[event.team.id].side.toLowerCase(),
            qualifiers: extractQualifierFlags(event),
            start: {
                x: utils.translateX(direction, event.x),
                y: utils.translateY(direction, event.y)
            }
        });
    }
    // TODO: plot corner awarded for outcome=1, calculate left/right corner position for the event (instead of using event's x/y)
}

function generateLandscape (context, event) {
    var player = event.player && context.players[event.player.id];
    var team = context.teams[event.team.id];
    var team_name = team ? team.name : 'Unknown team';
    var previous1GoalLeads = R.filter ( function ( scores ) {
        return ( scores[side] - scores[otherSide] ) === 1;
    }, context.scoreHistory ).length;

    var previous2GoalLeads = R.filter ( function ( scores ) {
        return ( scores[side] - scores[otherSide] ) === 2;
    }, context.scoreHistory ).length;

    var side = context.teams[event.team.id].side.toLowerCase();
    var otherSide = ( side === 'home' ? 'away' : 'home' );
    var camelCase = function ( string ) {
        return 'is' + string.split ( ' ' ).map ( function ( word ) {
            return word.charAt ( 0 ).toUpperCase () + word.slice ( 1 );
        } ).join ( '' );
    };

    context.particle.title = event.team.name + ' passes';

    if ( context.previousTypeAttr ) {
        context.particle.data[context.previousTypeAttr] = false;
    }

    if ( camelCase ( event.type.name ) ) {
        context.previousTypeAttr = camelCase ( event.type.name );
        context.particle.data[camelCase ( event.type.name )] = true;
    }

    context.particle.data.isGoal = false;

    if (event.type.name === 'Goal') {
        context.particle.data.isGoal = true;

        if (event.qualifiers['own-goal']) {
            context.scores[otherSide]++;
            context.particle.data.ownGoal = true;
        } else {
            context.scores[side]++;
            context.particle.data.ownGoal = false;
        }

        context.particle.data.homeScore = context.scores.home;
        context.particle.data.awayScore = context.scores.away;

        context.particle.title = [
            event.qualifiers['own-goal'] ? 'Own goal:' : 'Goal:',
            [
                [ context.particle.data.homeTeamName, context.scores.home ].join ( ' ' ),
                [ context.scores.away, context.particle.data.awayTeamName ].join ( ' ' )
            ].join ( ' - ' ),
            [ '(', ( player ? player.name : '' ), ', ', utils.getFormattedTime(event), ')' ].join ( '' )
        ].join ( ' ' );
    } else if (utils.endsChain (event)) {
        context.particle.title = [
            event.type.name,
            [ '(', ( player ? player.name : '' ), ', ', utils.getFormattedTime(event), ')' ].join ( '' )
        ].join ( ' ' );
    }

    context.particle.data.goalScorerPlayerName = ( (event.type.name === 'Goal') && player.name ) ? player.name: '';
    context.particle.data.goalScorerTeamName = (event.type.name === 'Goal') ? (side === 'home' ? context.particle.data.homeTeamName : context.particle.data.awayTeamName) : '';
    context.particle.data.otherTeamName = (event.type.name === 'Goal') ? (otherSide === 'home' ? context.particle.data.homeTeamName : context.particle.data.awayTeamName) : '';

    context.particle.data.isEqualiser = (event.type.name === 'Goal') && (context.scores.home === context.scores.away);
    context.particle.data.isOpener = (event.type.name === 'Goal') && (context.scores[side] === 1) && (context.scores[otherSide] === 0);
    context.particle.data.is20Lead = (event.type.name === 'Goal') && (context.scores[side] === 2) && (context.scores[otherSide] === 0);
    context.particle.data.is30Lead = (event.type.name === 'Goal') && (context.scores[side] === 3) && (context.scores[otherSide] === 0);
    context.particle.data.is60Lead = (event.type.name === 'Goal') && (context.scores[side] === 6) && (context.scores[otherSide] === 0);
    context.particle.data.is70Lead = (event.type.name === 'Goal') && (context.scores[side] === 7) && (context.scores[otherSide] === 0);
    context.particle.data.is1GoalLead = (event.type.name === 'Goal') && ((context.scores[side] - context.scores[otherSide]) === 1);
    context.particle.data.is2GoalLead = (event.type.name === 'Goal') && ((context.scores[side] - context.scores[otherSide]) === 2);
    context.particle.data.is2nd1GoalLead = (event.type.name === 'Goal') && (context.particle.data.is1GoalLead && (previous1GoalLeads === 1));
    context.particle.data.is2nd2GoalLead = (event.type.name === 'Goal') && (context.particle.data.is2GoalLead && (previous2GoalLeads === 1));
    context.particle.data.isAnschlussTreffer = (event.type.name === 'Goal') && ((context.scores[side] === 1) && (context.scores[otherSide] === 2) && (event.min < 75));
    context.particle.data.isGoalPulledBack = (event.type.name === 'Goal') && ((context.scores[otherSide] - context.scores[side]) === 2);
    context.particle.data.isConsolationGoal = (event.type.name === 'Goal') && (((context.scores[otherSide] - context.scores[side]) > 2) && (event.min > 75));

    context.particle.data['current-period'] = utils.setTimePeriodForRules(event, context);
    context.particle.data.isEarly = (event.min < 15);
    context.particle.data.isLate = (event.min > 45);
    context.particle.data.isLateLate = (event.min > 80);
    context.particle.data.isJustBeforeHalftime = (event.min > 42) && (context.particle.data['current-period'].id === 1);

    if (event.type.name === 'Goal') {
        context.scoreHistory.push ( R.clone ( context.scores ) );
    }
}

exports.initial = function (config) {
    var findTeam = utils.chooseTeamName (R.values(config.teams));

    return {
        particle: {
            title: 'Passes',
            type: exports.particleType,
            assemblerType: assemblerType,
            version: config.version,
            uniqueId: [ assemblerType, 0 ].join ( '-' ),
            data: {
                'current-period': {
                    id: -1,
                    'is-first-event': false,
                    'is-start-half-time': false
                },
                animated: false,
                showDirectionOfAttack: true,
                homeTeamName: findTeam ('home'),
                awayTeamName: findTeam ('away'),
                homeScore: 0,
                awayScore: 0,
                goalScorerPlayerName: '',
                goalScorerTeamName: '',
                otherTeamName: '',
                isEarly: false,
                isLate: false,
                isLateLate: false,
                isJustBeforeHalftime: false,
                isEqualiser: false,
                isOpener: false,
                is20Lead: false,
                is30Lead: false,
                is60Lead: false,
                is70Lead: false,
                is1GoalLead: false,
                is2GoalLead: false,
                is2nd1GoalLead: false,
                is2nd2GoalLead: false,
                isAnschlussTreffer: false,
                isGoalPulledBack: false,
                isConsolationGoal: false,
		        events: []
            }
        },
        scoreHistory: [],
        interrupted: false,
        teams: config.teams,
	    currentSide: null, // side the current event chain is for
        players: {},
	    scores: {
            home: 0,
            away: 0
        },
        meta: {
            type: 'passes',
            publishTags: [ 'reducer:rbb-passes-on-pitch' ],
            generatedTimeStamp: null,
            eventTimeStamp: null            
        },
        gameTime: null
    };
};

exports.reducer = function (context, event) {
    if (utils.excludeFromChain(event)) {
	    return context;
    }
    // update team and player metadata
    utils.extendTeamData(context, event);
    utils.extendPlayerData(context, event);

    if (utils.breaksChain(event)) {
	    // mark chain as broken, do not add this event to chain
	    context.interrupted = 'always';
    }
    else if (utils.endsChain(event)) {
	    // first, add this event to the chain
	    extendChain(context, event);
	    // then mark as broken
	    context.interrupted = 'always';
    }
    else if (utils.maybeEndsChain(event)) {
	    // first, add this event to the chain
	    extendChain(context, event);
	    // then mark as maybe broken
	    context.interrupted = 'maybe';
    }
    else if (utils.isGenericPassEvent(event)) {
	    extendChain(context, event);
    }
    
    generateLandscape(context, event);

    context.gameTime = utils.createGameTimeObject(event, context.gameTime);
    // opta recorded time
    context.meta.eventTimeStamp = utils.getOptaEventTimeStamp(event);
    
    return context;
};
