var debug = false;

var R = require ( 'ramda' );
var utils = require ( './utils' );

var log = debug ? console.log : function () {};

exports.particleType = 'rbb-single-pitch-map';

exports.assemblerType = 'rbb-event-chains';

exports.initial = function ( config ) {
    var findTeam = function ( teams, side ) {
        return R.find ( function ( team ) {
            return team.side === side;
        }, teams );
    };

    return {
        particle: {
            version: config.version,
            type: exports.particleType,
            assemblerType: exports.assemblerType,
            title: utils.chooseTeamName ( R.values ( config.teams ), 'Home' ) + ' vs ' + utils.chooseTeamName ( R.values ( config.teams ), 'Away' ),
            data: {
                animated: true,
                showDirectionOfAttack: false,
                homeTeamName: utils.chooseTeamName ( R.values ( config.teams ), 'Home' ),
                awayTeamName: utils.chooseTeamName ( R.values ( config.teams ), 'Away' ),
                homeScore: 0,
                awayScore: 0,
                ownGoal: false,
                events: []
            }
        },
        scoreHistory: [],
        teams: config.teams,
        events: []
    };
};

exports.reducer = function ( context, event ) {
    var ignoreEvents = [
        'Good skill',
        'Keeper Sweeper',
        'Condition change',
        'Aerial',
        'Deleted event',
        'Offside provoked',
        //'Card',
        'Penalty faced',
        'Collection End',
        'Player Off',
        'Player retired',
        'Player returns',
        'Goalkeeper becomes player',
        'Player becomes goalkeeper',
        'Official change',
        'Player changed position',
        'Player changed Jersey number',
        'Collection End',
        'Temp_Goal',
        'Temp_Attempt',
        'Rescinded card',
        'Cross not claimed',
        'Shield ball opp',
        'Contentious referee decision',
        'Injury Time Announcement',
        'Chance missed',
        'Temp_Save',
        'Resume',
        '50/50',
        'Failed to Block',
        'Coach Setup',
        'Error'
    ];

    var breakOnEvents = [
        'Out',
        'Offside Pass',
        'Caught Offside',
        'Goal',
        'Foul',
        'Foul throw-in',
        //'Card',
        //'Corner Awarded'
    ];

    var ignoreAndBreakEvents = [
        'End',
        //'Corner Awarded',
        'Start delay',
        'End delay',
        'Formation change',
        'Referee Drop Ball'
    ];

    var alterAndIgnoreEvents = [
        'Card'
    ];

    var addIgnoreAndBreakEvents = [
        'Corner Awarded'
    ];

    var removeNegativeOutcomeEvents = [
        'Ball touch',
        'Dispossessed',
        'Tackle',
        'Take On',
        'Challenge',
        'Corner Awarded'
    ];

    var removePositiveOutcomeEvents = [
        'Out',
        'Foul'
    ];

    var drawnButNotCountedEvents = [
        'Ball recovery',
        'Save'
    ];

    var shotEvents = [
        'Miss',
        'Post',
        'Goal',
        'Attempt Saved',
        'Save'
    ];

    var particleDataPath = [ 'particle', 'data' ];
    var particleAttackLeftPath = R.concat ( particleDataPath, [ 'attack-left' ] );
    var particleActionLeftPath = R.concat ( particleDataPath, [ 'action-left' ] );
    var particleEventsPath = R.concat ( particleDataPath, [ 'events' ] );
    var particleGoalkeepersPath = R.concat ( particleDataPath, [ 'goalkeepers' ] );

    var goalkeepersPath = [ 'goalkeepers' ];
    var eventsPath = [ 'events' ];
    var breakIndicatorPath = [ 'break' ];
    var eventBufferPath = [ 'eventBuffer' ];
    var currSidePath = [ 'currentSide' ];

    var mapKey = function ( key, func ) {
        return R.converge ( R.merge, [
            R.identity,
            R.compose ( R.fromPairs, R.of, R.concat ( [ key ] ), func, R.prop ( key ) )
        ] );
    };

    var otherDirection = function ( direction ) {
        if ( direction === 'Left to Right' ) {
            return 'Right to Left';
        }

        return 'Left to Right';
    };

    var otherSide = function ( side ) {
        return ( side === 'home' ) ? 'away' : 'home';
    };

    var findPlayer = function ( playerId ) {
        return R.find ( function ( player ) {
            return player.id === playerId;
        }, R.concat ( R.values ( context.teams )[0].players, R.values ( context.teams )[1].players ) );
    };

    var eventType = function ( event ) {
        if ( event.type.name === 'Goal' ) {
            return 'goal';
        }

        if ( event.type.name === 'Post' ) {
            return 'post';
        }

        if ( R.contains ( event.type.name, shotEvents ) ) {
            return 'pass';
        }

        if ( R.has ( 'pass-end-x', event.qualifiers ) && R.has ( 'pass-end-y', event.qualifiers ) ) {
            return 'pass';
        }

        return 'position';
    };

    var eventEnd = function ( event ) {
        if ( R.contains ( event.type.name, shotEvents ) ) {
            if ( event.type.name === 'Save' ) {
                return undefined;
            }

            if ( R.has ( 'blocked-x-co-ordinate', event.qualifiers ) ) {
                return {
                    x: event.qualifiers['blocked-x-co-ordinate'],
                    y: event.qualifiers['blocked-y-co-ordinate']
                };
            }

            return {
                x: utils.translateX ( ( R.has ( 'own-goal', event.qualifiers ) ? otherDirection : R.identity )( context.teams[event.team.id].direction ), 100 ),
                y: event.qualifiers['goal-mouth-y-co-ordinate']
            };
        }

        if ( R.has ( 'pass-end-x', event.qualifiers ) && R.has ( 'pass-end-y', event.qualifiers ) ) {
            return {
                x: event.qualifiers['pass-end-x'],
                y: event.qualifiers['pass-end-y']
            };
        }

        return undefined;
    };

    var eventStatic = function ( event ) {
        return ( event.type.name === 'Save' ) || ! ( R.contains ( event.type.name, shotEvents ) || ( R.has ( 'pass-end-x', event.qualifiers ) && R.has ( 'pass-end-y', event.qualifiers ) ) );
    };

    var transformEvent = R.compose (
        function ( event ) {
            var player = event.player ? findPlayer ( event.player.id ) : undefined;
            var playerName = ( player && player.name ) ? ( player.name.known || player.name.first + ' ' + player.name.last ) : '';

            return {
                type: eventType ( event ),
                desc: event.type.name,
                static: eventStatic ( event ),
                side: R.toLower ( event.team.side ),
                qualifiers: event.qualifiers,
                player: {
                    name: playerName,
                    jersey: ( player && player.jersey ) ? player.jersey : '',
                    display: true
                },
                start: {
                    x: event.x,
                    y: event.y
                },
                end: eventEnd ( event ),
                direction: ( event.team ) ? context.teams[event.team.id].direction : ''
            };
        },
        mapKey ( 'qualifiers', function ( qualifiers ) {
            var translatedAttributes = {
                'pass-end-x': utils.translateX,
                'pass-end-y': utils.translateY,
                'blocked-x-co-ordinate': utils.translateX,
                'blocked-y-co-ordinate': utils.translateY,
                'gk-x-coordinate': utils.translateX,
                'gk-y-coordinate': utils.translateY,
                'goal-mouth-y-co-ordinate': utils.translateY
            };

            var translatedDir = {
                'pass-end-x': R.identity,
                'pass-end-y': R.identity,
                'blocked-x-co-ordinate': R.identity,
                'blocked-y-co-ordinate': R.identity,
                'gk-x-coordinate': R.always ( R.has ( 'own-goal', event.qualifiers ) ? 'Right to Left' : 'Left to Right' ),
                'gk-y-coordinate': R.always ( R.has ( 'own-goal', event.qualifiers ) ? 'Right to Left' : 'Left to Right' ),
                'goal-mouth-y-co-ordinate': R.identity
            };

            if ( event.team && event.team.id ) {
                return R.reduce ( function ( qualifiers, qualifierKey ) {
                    if ( R.has ( qualifierKey, qualifiers ) ) {
                        return R.merge ( qualifiers, R.fromPairs ( [ [
                            qualifierKey, 
                            translatedAttributes[qualifierKey] ( translatedDir[qualifierKey] ( context.teams[event.team.id].direction ), qualifiers[qualifierKey] )
                        ] ] ) );
                    }

                    return qualifiers;
                }, qualifiers, R.keys ( translatedAttributes ) );
            }

            return R.omit ( R.keys ( translatedAttributes ), qualifiers );
        } ),
        mapKey ( 'qualifiers', R.map ( function ( qualifier ) {
            return qualifier.value || true;
        } ) ),
        mapKey ( 'player', function ( player ) {
            if ( player ) {
                return findPlayer ( player.id );
            }

            return undefined;
        } ),
        mapKey ( 'team', function ( team ) {
            if ( team ) {
                return R.omit ( [ 'players' ], R.find ( function ( t ) {
                    return t.id === team.id;
                }, R.values ( context.teams ) ) );
            }

            return undefined;
        } ),
        mapKey ( 'x', function ( x ) {
            if ( x && event.team && event.team.id ) {
                return utils.translateX ( context.teams[event.team.id].direction, x );
            }

            return undefined;
        } ),
        mapKey ( 'y', function ( y ) {
            if ( y && event.team && event.team.id ) {
                return utils.translateY ( context.teams[event.team.id].direction, y );
            }

            return undefined;
        } ),
        R.pick ( [
            'player',
            'team',
            'id',
            'type',
            'period_id',
            'min',
            'sec',
            'outcome',
            'x',
            'y',
            'qualifiers'
        ] )
    );

    var txEvent = transformEvent ( event );

    var sideSwitch = function ( context, txEvent ) {
        var last3sides = R.concat ( [ txEvent.side ], R.map ( R.prop ( 'side' ), R.path ( eventBufferPath, context ) || [] ) );

        return last3sides[0] !== R.path ( currSidePath, context ) && last3sides[0] === last3sides[1] && last3sides[1] !== last3sides[2];
    };

    var playerIcon = function ( eventType, qualifiers ) {
        if ( eventType === 'Out' ) {
            return { icon: 'out', display: true };
        }

        if ( eventType === 'Corner Awarded' ) {
            if ( R.has ( 'left', qualifiers ) ) {
                return { icon: 'corner-left', display: true };
            }

            if ( R.has ( 'right', qualifiers ) ) {
                return { icon: 'corner-right', display: true };
            }

            return { icon: 'corner', display: true };
        }

        if ( eventType === 'Card' ) {
            if ( R.has ( 'second-yellow', qualifiers ) ) {
                return { icon: 'second-yellow-card', display: true };
            }

            if ( R.has ( 'red-card', qualifiers ) ) {
                return { icon: 'red-card', display: true };
            }

            return { icon: 'yellow-card', display: true };
        }

        return {};
    };

    var populateParticle = function ( context ) {
        var ctx = R.reduce ( function ( ctx, event ) {
            var auxiliaryEvents = ( event.desc === 'Goal' ) ? [
                {
                    type: 'pass',
                    desc: 'Goalkeeper',
                    static: true,
                    side: R.has ( 'own-goal', event.qualifiers ) ? event.side : otherSide ( event.side ),
                    qualifiers: {},
                    player: {
                        name: '',
                        jersey: 'GK',
                        display: true
                    },
                    start: {
                        x: event.qualifiers['gk-x-coordinate'],
                        y: event.qualifiers['gk-y-coordinate']
                    }
                }
            ] : [];

            if ( event.desc === 'Save' || ! ctx.previousEvent ) {
                return {
                    previousEvent: event,
                    events: R.concat ( R.concat ( ctx.events, [ event ] ), auxiliaryEvents )
                };
            }

            var displayOnPreviousEnd = ! ( R.contains ( ctx.previousEvent.desc, shotEvents ) || ( R.last ( ctx.events ).player.name === event.player.name ) || ( R.path ( [ 'previousEvent', 'static' ], ctx ) && event.static ) );

            var displayOnStart = R.contains ( ctx.previousEvent.desc, shotEvents ) || ( R.path ( [ 'previousEvent', 'static' ], ctx ) && event.static );

            return {
                previousEvent: event,
                events: R.concat ( R.concat ( ctx.events, [ {
                    type: ( R.contains ( ctx.previousEvent.desc, shotEvents ) || ( R.path ( [ 'previousEvent', 'static' ], ctx ) && event.static ) ) ? 'pass' : 'dribble',
                    desc: ( R.contains ( ctx.previousEvent.desc, shotEvents ) || ( R.path ( [ 'previousEvent', 'static' ], ctx ) && event.static ) ) ? 'Implied pass' : 'Dribble',
                    static: false,
                    side: event.side,
                    qualifiers: {},
                    player: R.merge ( event.player, {
                        display: displayOnPreviousEnd
                    } ),
                    start: {
                        x: ctx.previousEvent.end ? ctx.previousEvent.end.x : ctx.previousEvent.start.x,
                        y: ctx.previousEvent.end ? ctx.previousEvent.end.y : ctx.previousEvent.start.y
                    },
                    end: {
                        x: event.start.x,
                        y: event.start.y
                    }
                }, R.merge ( event, {
                    player: R.merge ( {
                        name: event.player.name,
                        jersey: event.player.jersey,
                        display: displayOnStart
                    }, playerIcon ( event.desc, event.qualifiers ) ) 
                } ) ] ), auxiliaryEvents )
            };
        }, { events: [] }, context.events || [] );

        var markLastX = R.curry ( function ( x, events ) {
            for ( var i = events.length - 1; i >= 0; i-- ) {
                events[i]['last-' + x] = i >= events.length - x || ( events[i].type === 'dribble' && events[i + 1]['last-' + x] === true );
            }
        } );

        markLastX ( 20, ctx.events );

        return R.assocPath ( particleEventsPath, ctx.events, R.assocPath (
            particleGoalkeepersPath,
            R.path ( goalkeepersPath, context ),
            context
        ) );
    };

    var addEvent = function ( context, event, txEvent ) {
        var calculateScore = function ( context, event, side ) {

            var scorerSide = R.toLower ( context.teams[event.team.id].side );
            var ownGoal = R.has ( 'own-goal', event.qualifiers );
            var scorePath = R.concat ( particleDataPath, [ side + 'Score' ] );

            return R.path ( scorePath, context ) + ( ( event.type.name === 'Goal' ) ? ( ( side === scorerSide ) ? ( ownGoal ? 0 : 1 ) : ( ownGoal ? 1 : 0 ) ) : 0 );
        };

        var elementsToAdd = [
            [ breakIndicatorPath, R.contains ( event.type.name, breakOnEvents ) ],
            [ eventsPath, R.concat (
                ( ! R.path ( breakIndicatorPath, context ) && ( sideSwitch ( context, txEvent ) ? R.slice ( 0, 1, R.path ( eventBufferPath, context ) || [] ) : R.path ( eventsPath, context ) ) ) || [],
                [ txEvent ]
            ) ],
            [ eventBufferPath, R.slice ( 0, 2, R.concat ( R.contains ( event.type.name, drawnButNotCountedEvents ) ? [] : [ txEvent ], R.path ( eventBufferPath, context ) || [] ) ) ],
            [ currSidePath, ( sideSwitch ( context, txEvent ) ? ( R.path ( currSidePath, context ) === 'home' ? 'away' : 'home' ) : R.path ( currSidePath, context ) ) ],
            [ particleAttackLeftPath, R.has ( 'own-goal', txEvent.qualifiers ) ? ( txEvent.direction === 'Left to Right' ) : ( txEvent.direction === 'Right to Left' ) ],
            [ particleActionLeftPath, R.has ( 'own-goal', txEvent.qualifiers ) ? ( txEvent.direction === 'Left to Right' ) : ( txEvent.direction === 'Right to Left' ) ],
            [ R.concat ( particleDataPath, [ 'goalScorerPlayerName' ] ), ( ( ( event.type.name === 'Goal' ) && txEvent.player ) ? txEvent.player.name: '' ) ],
            [ R.concat ( particleDataPath, [ 'goalScorerTeamName' ] ), ( ( ( event.type.name === 'Goal' ) && event.team ) ? ( context.teams[event.team.id].side === 'Home' ? context.particle.data.homeTeamName : context.particle.data.awayTeamName ) : '' ) ],
            [ R.concat ( particleDataPath, [ 'otherTeamName' ] ), ( ( ( event.type.name === 'Goal' ) && event.team ) ? ( context.teams[event.team.id].side === 'Home' ? context.particle.data.awayTeamName : context.particle.data.homeTeamName ) : '' ) ],
            [ R.concat ( particleDataPath, [ 'homeScore' ] ), calculateScore ( context, event, 'home' ) ],
            [ R.concat ( particleDataPath, [ 'awayScore' ] ), calculateScore ( context, event, 'away' ) ],
            [ [ 'particle', 'title' ], ( ( event.type.name !== 'Goal' ) ? ( event.type.name + ': ' ) : '' ) + [
                //event.qualifiers['own-goal'] ? 'Own goal:' : ( event.type.name + ':' ),
                [
                    [
                        R.path ( R.concat ( particleDataPath, 'homeTeamName' ), context ),
                        calculateScore ( context, event, 'home' )
                    ].join ( ' ' ),
                    [
                        calculateScore ( context, event, 'away' ),
                        R.path ( R.concat ( particleDataPath, 'awayTeamName' ), context )
                    ].join ( ' ' )
                ].join ( ' - ' ),
                [ '(', ( txEvent.player ? txEvent.player.name : 'Unknown player' ), ', ', utils.getFormattedTime(event), ')' ].join ( '' )
            ].join ( ' ' ) ],
            [ R.concat ( particleDataPath, [ 'ownGoal' ] ), R.has ( 'own-goal', event.qualifiers ) ] 
        ];

        return populateParticle ( R.reduce ( function ( context, elementPair ) {
            return R.assocPath ( elementPair[0], elementPair[1], context );
        }, context, elementsToAdd ) );
    };

    log ( event.type.name );

    log ( event.qualifiers );

    if ( event.type.name === 'Team set up' ) {
        var playerJerseyNumbers = R.fromPairs ( R.zip (
            event.qualifiers.involved.value.split ( ', ' ),
            event.qualifiers['jersey-number'].value.split ( ', ' )
        ) );

        var playerPositions = R.fromPairs ( R.zip (
            event.qualifiers.involved.value.split ( ', ' ),
            event.qualifiers['player-position'].value.split ( ', ' )
        ) );

        return R.assocPath ( [ 'teams', event.team.id, 'players' ], R.map ( function ( player ) {
            if ( playerPositions[player.id] === '1' ) {
                return R.assocPath ( [ 'jersey' ], 'GK', player );
            }
            return R.assocPath ( [ 'jersey' ], playerJerseyNumbers[player.id] || '', player );
        }, R.path ( [ 'teams', event.team.id, 'players' ], context ) ), context );
    }

    if ( event.type.name === 'Player on' ) {
        if ( R.path ( [ 'qualifiers', 'player-position', 'value' ], event ) === 'Goalkeeper' ) {
            return R.assocPath ( [ 'teams', event.team.id, 'players' ], R.map ( function ( player ) {
                if ( event.player.id === player.id ) {
                    return R.assocPath ( [ 'jersey' ], 'GK', player );
                }
                return player;
            }, R.path ( [ 'teams', event.team.id, 'players' ], context ) ), context );
        }
        return context;
    }

    if ( event.type.name === 'Start' ) {
        log ( 'start - direction: ' + event.qualifiers['direction-of-play'].value );
        return R.assocPath ( [ 'teams', event.team.id, 'direction' ], event.qualifiers['direction-of-play'].value, context );
    }

    if ( R.contains ( event.type.name, ignoreAndBreakEvents ) ) {
        log ( 'ignored and broken' );
        return R.assocPath ( breakIndicatorPath, true, context );
    }

    if ( R.contains ( event.type.name, ignoreEvents ) ) {
        log ( 'ignored' );
        return context;
    }

    if ( R.contains ( event.type.name, removeNegativeOutcomeEvents ) && ( event.outcome === 0 ) ) {
        log ( 'negative outcome with outcome ' + event.outcome );
        return context;
    }

    if ( R.contains ( event.type.name, removePositiveOutcomeEvents ) && ( event.outcome === 1 ) ) {
        log ( 'positive outcome with outcome ' + event.outcome );
        return context;
    }

    if ( R.contains ( event.type.name, alterAndIgnoreEvents ) ) {
        log ( 'altered and ignored' );
        return R.merge ( context, {
            particle: R.merge ( context.particle, {
                title: R.concat ( [ event.type.name ], R.tail ( R.split ( ':', context.particle.title ) ) ).join ( ':' ),
                data: R.merge ( context.particle.data, {
                    events: ( context.particle.data.events.length === 0 ) ? [] : ( R.concat (
                        ( context.particle.data.events.length === 1 ) ? [] : R.init ( context.particle.data.events ),
                        R.of ( R.merge ( R.last ( context.particle.data.events ), {
                            player: R.merge ( R.last ( context.particle.data.events ).player, playerIcon ( event.type.name, event.qualifiers ) )
                        } ) )
                    ) )
                } )
            } )
        } );
    }

    if ( R.contains ( event.type.name, addIgnoreAndBreakEvents ) ) {
        log ( 'altered and ignored' );

        return R.assocPath ( breakIndicatorPath, true, R.merge ( context, {
            particle: R.merge ( context.particle, {
                title: R.concat ( [ event.type.name ], R.tail ( R.split ( ':', context.particle.title ) ) ).join ( ':' ),
                data: R.merge ( context.particle.data, {
                    events: R.concat ( context.particle.data.events, [ R.merge ( txEvent, {
                        player: R.merge ( txEvent.player, playerIcon ( event.type.name, event.qualifiers ) ),
                        'last-20': true
                    } ) ] )
                } )
            } )
        } ) );
    }

    if ( debug && context.break ) {
        log ( 'previous event was breaking - start a new chain' );
    }

    if ( debug && R.contains ( event.type.name, breakOnEvents ) ) {
        log ( 'breaking event' );
    }

    if ( debug && sideSwitch ( context, txEvent ) ) {
        log ( 'sides switched' );
    }

    return addEvent ( context, event, txEvent );
};
