var utils = require('./utils');

var assemblerType = 'rbb-ineffective-possession';

exports.particleType = 'rbb-comparison';

exports.initial = function (config) {
    return {
        particle: {
            title: 'Possession over last 10 minutes',
            type: exports.particleType,
            assemblerType: assemblerType,
            version: config.version,
            uniqueId: [ assemblerType, 0 ].join ( '-' ),
            data: {}
        },
        passchains: [],
        interrupted: false,
        teams: config.teams,
        meta: {
            type: 'inneffective-possession',
            publishTags: [ 'reducer:rbb-inneffective-possession' ],
            generatedTimeStamp: null,
            eventTimeStamp: null          
        },
        gameTime: null
    };
};

function moveTimeWindow (context, now_secs) {
    // trim events to only include last 10 minutes
    context.passchains = context.passchains.map(function (chain, i) {
        return chain.filter(function (event) {
            return event.timestamp >= (now_secs - 600);
        });
    });
    context.passchains = context.passchains.filter(function (chain) {
        return chain.length > 0;
    });
    if (context.passchains.length === 0) {
        context.passchains = [[]];
    }
}

function resetPassChain(context) {
    context.passchains = [];
    nextPassChain(context);
}

function nextPassChain(context) {
    context.passchains.push([]);
    context.interrupted = false;
}

function toPercentages(a, b, round, min) {
    var total = a + b;
    if (total === 0) {
        return ['50%', '50%'];
    }
    var ap = a/total*100;
    var bp = b/total*100;
    if (round) {
        ap = Math.round(ap);
        bp = Math.round(bp);
    }
    if (min) {
        if (ap < min) {
            bp -= min - ap;
            ap = min;
        }
        if (bp < min) {
            ap -= min - bp;
            bp = min;
        }
    }
    return [ap + '%', bp + '%'];
}

exports.reducer = function (context, event) {
    utils.extendTeamData(context, event);

    var team;
    if (event.timestamp) {
        moveTimeWindow(context, event.timestamp.unix);
    }
    if (event.type.name === 'Team set up') {
        team = context.teams[event.team.id];
    }
    else if (event.type.name === 'Start') {
        context.teams[event.team.id].direction = event.qualifiers['direction-of-play'].value;
        resetPassChain(context);
    }
    else if (utils.isGenericPassEvent(event)) {
        if (utils.isPassChainStart(event, context.last_team) || context.interrupted) {
            nextPassChain(context);
        }
        team = context.teams[event.team.id];
        var chain = context.passchains[context.passchains.length-1];
        chain.push({
            type: event.qualifiers['throw-in'] ? 'throw-in': 'pass',
            team: event.team.id,
            timestamp: event.timestamp.unix
        });
        context.last_x = event.qualifiers['pass-end-x'].value;
        context.last_y = event.qualifiers['pass-end-y'].value;
        context.last_team = event.team.id;
        context.particle.uniqueId = [ assemblerType, event.id ].join ( '-' );
    }
    else if (event.type.name === 'Miss') {
        team = context.teams[event.team.id];
        var chain = context.passchains[context.passchains.length-1];
        chain.push({
            type: 'miss',
            team: event.team.id,
            timestamp: event.timestamp.unix
        });
        context.particle.uniqueId = [ assemblerType, event.id ].join ( '-' );
    }
    else if (utils.isPassChainEnd(event, context.last_team)) {
        context.interrupted = true;
    }

    // update stats
    var team_ids = [];
    // build team ids this way to preserve colours for home/away as
    // used in other particles
    for (var id in context.teams) {
        if (context.teams[id].side.toLowerCase() === 'home') {
            team_ids[0] = id;
        }
        else {
            team_ids[1] = id;
        }
    }
    if (Object.keys(context.teams).length == 2) {

        var possession_time = {};
        var passes = {};
        var misses = {};

        team_ids.forEach(function (id) {
            possession_time[id] = 0;
            passes[id] = 0;
            misses[id] = 0;
            context.passchains.forEach(function (chain) {
                if (chain.length) {
                    if (chain[0].team == id) {
                        var chain_start = chain[0].timestamp;
                        var chain_end = chain[chain.length-1].timestamp;
                        var chain_duration = chain_end - chain_start;
                        possession_time[id] += chain_duration;
                        var chain_passes = chain.filter(function (event) {
                            return event.type === 'pass';
                        });
                        passes[id] += chain_passes.length;
                        var chain_misses = chain.filter(function (event) {
                            return event.type === 'miss';
                        });
                        misses[id] += chain_misses.length;
                    }
                }
            });
        });

        context.particle.data = {
            duel: {
                opponents: [
                    {
                        name: context.teams[team_ids[0]].name,
                        url: ''
                    },
                    {
                        name: context.teams[team_ids[1]].name,
                        url: ''
                    }
                ],
                stats: [
                    {
                        type: 'Possession',
                        values: toPercentages(
                            possession_time[team_ids[0]],
                            possession_time[team_ids[1]],
                            true
                        ),
                        sizes: toPercentages(
                            possession_time[team_ids[0]],
                            possession_time[team_ids[1]],
                            false,
                            10
                        )
                    },
                    {
                        type: 'Passes',
                        values: [
                            passes[team_ids[0]],
                            passes[team_ids[1]]
                        ],
                        sizes: toPercentages(
                            passes[team_ids[0]],
                            passes[team_ids[1]],
                            false,
                            10
                        )
                    },
                    {
                        type: 'Missed shots',
                        values: [
                            misses[team_ids[0]],
                            misses[team_ids[1]]
                        ],
                        sizes: toPercentages(
                            misses[team_ids[0]],
                            misses[team_ids[1]],
                            false,
                            10
                        )
                    }
                ]
            }
        };
        context.gameTime = utils.createGameTimeObject(event, context.gameTime);
        // opta recorded time
        context.meta.eventTimeStamp = utils.getOptaEventTimeStamp(event);
    }
    return context;
};
