var dashboard = require('./rbb-dashboard'),
    utils = require('./utils'),
    R = require ( 'ramda' );


var assemblerType = 'rbb-possesion';

exports.particleType = 'rbb-comparison';

exports.initial = function (config) {

    var findTeam = utils.chooseTeamName (R.values(config.teams));

    return {
        particle: {
            title: 'Possession over last 10 minutes',
            type: exports.particleType,
            assemblerType: assemblerType,
            version: config.version,
            uniqueId: [ assemblerType, 0 ].join ( '-' ),
            data: {
                timeWindow: '',
                timelineEvents: [],        		        		
            }
        },
        passchains: [],
        interrupted: false,
        teams: config.teams,
        meta: {
            type: 'possession',
            publishTags: [ 'reducer:rbb-possession' ],
            generatedTimeStamp: null,
            eventTimeStamp: null           
        },
        gameTime: null,
        homeTeamName: findTeam ('home'),
        awayTeamName: findTeam ('away')
    };
};

var convertToTimelinePct = R.compose( R.multiply(100), R.min(1) );

function moveTimeWindow (context, now_secs) {
    // trim events to only include last 10 minutes
    context.passchains = context.passchains.map(function (chain, i) {
        return chain.filter(function (event) {
            return event.timestamp >= (now_secs - 600);
        });
    });
    context.passchains = context.passchains.filter(function (chain) {
        return chain.length > 0;
    });
    if (context.passchains.length === 0) {
        context.passchains = [[]];
    }
    // update timeline events too
    context.particle.data.timelineEvents = context.particle.data.timelineEvents.filter(function (event) {
            return event.timestamp >= (now_secs - 600);
    });
    // update timeline item positions
    context.particle.data.timelineEvents =  R.map( function( event ) {
        event.timeline.pct.x = convertToTimelinePct( (event.timestamp - (now_secs - 600) ) / 600 );
        return event;
    }, context.particle.data.timelineEvents );

}

function resetPassChain(context) {
    context.passchains = [];
    nextPassChain(context);
}

function nextPassChain(context) {
    context.passchains.push([]);
    context.interrupted = false;
}

function toPercentages(a, b, round, min) {
    var total = a + b;
    if (total === 0) {
        return ['50%', '50%'];
    }
    var ap = a/total*100;
    var bp = b/total*100;
    if (round) {
        ap = Math.round(ap);
        bp = Math.round(bp);
    }
    if (min) {
        if (ap < min) {
            bp -= min - ap;
            ap = min;
        }
        if (bp < min) {
            ap -= min - bp;
            bp = min;
        }
    }
    return [ap + '%', bp + '%'];
}

function addToChain(context, event) {
    var chain = context.passchains[context.passchains.length-1];
    if (chain.length && context.maybe_interrupted) {
	// have the last two events been from the opposing team to the
	// one that started the chain?
	var chain_starter = chain[0].team;
	var last_team = chain[chain.length-1].team;
	if (event.team.id == last_team && last_team != chain_starter) {
	    // it's a real interruption
	    nextPassChain(context);
	    return addToChain(context, event);
	}
	else {
	    // keep the chain going
	    context.maybe_interrupted = false;
	}
    }
    else if (chain.length && chain[chain.length-1].team !== event.team.id) {
	// swapped sides without knowing why
	nextPassChain(context);
	return addToChain(context, event);
    }
    chain.push({
	team: event.team.id,
	timestamp: event.timestamp.unix
    });
}

exports.reducer = function (context, event) {
    utils.extendTeamData(context, event);
    
    if (event.timestamp) {
        moveTimeWindow(context, event.timestamp.unix);
    }
    if (event.type.name === 'Start') {
        resetPassChain(context);
    }
    
    if (utils.excludeFromChain(event)) {
	// do nothing to the passchains
    }
    else if (utils.breaksChain(event)) {
	nextPassChain(context);
    }
    else if (utils.endsChain(event)) {
	addToChain(context, event);
	nextPassChain(context);
    }
    else if (utils.maybeEndsChain(event)) {
	context.maybe_interrupted = true;
	addToChain(context, event);
    }
    else if (utils.isGenericPassEvent(event)) {
	addToChain(context, event);
    }

    // update stats
    var team_ids = [];
    // build team ids this way to preserve colours for home/away as
    // used in other particles
    for (var id in context.teams) {
        if (context.teams[id].side.toLowerCase() === 'home') {
            team_ids[0] = id;
        }
        else {
            team_ids[1] = id;
        }
    }
    
    if (Object.keys(context.teams).length == 2) {
        var possession_time = {};
        var misses = {};
	
        team_ids.forEach(function (id) {
            possession_time[id] = 0;
            context.passchains.forEach(function (chain) {
                if (chain.length) {
                    if (chain[0].team == id) {
                        var chain_start = chain[0].timestamp;
                        var chain_end = chain[chain.length-1].timestamp;
                        var chain_duration = chain_end - chain_start;
                        possession_time[id] += chain_duration;
                    }
                }
            });
        });
	
        context.particle.data.duel = {
            opponents: [
                {name: context.homeTeamName, url: ''},
                {name: context.awayTeamName, url: ''}
            ],
            stats: [
                {
                    type: 'Possession',
                    hideTitle: true,
                    values: toPercentages(
                        possession_time[team_ids[0]],
                        possession_time[team_ids[1]],
                        true
                    ),
                    sizes: toPercentages(
                        possession_time[team_ids[0]],
                        possession_time[team_ids[1]],
                        false,
                        10
                    )
                }
            ]
        };
	
        context.gameTime = utils.createGameTimeObject(event, context.gameTime);
        // opta recorded time
        context.meta.eventTimeStamp = utils.getOptaEventTimeStamp(event);

        if (event.min) {
            context.particle.data.timeWindow = [ [
                utils.getSemiFormattedTime ( R.assocPath ( [ 'min' ], Math.max(event.min - 10, 0), event ) ),
                utils.getSemiFormattedTime ( event )
            ].join ( ' - ' ), 'min' ].join ( ' ' );
        }
    }

    // data for timeline
    var side = context.teams[event.team.id].side.toLowerCase();    
    
    // handle the events that belong in the timeline
    if( dashboard.isTimelineEvent(event) && dashboard.isGamePeriodEvent( event ) ) {
        var type = dashboard.slugifyEvent( event ) ;

        var timelineEvent = {
            side: side,
            type: type,
            is: {
                home: R.equals( side, 'home' ),
                goal: dashboard.isGoal(event),
                ownGoal: R.has('own-goal', event.qualifiers),
                card: dashboard.isCard(event),
                marker: dashboard.isMarkerEvent(event)
            },
            min: event.min,
            sec: event.sec,
            period_id: event.period_id,
            card: dashboard.getCardType(event),
            timestamp: event.timestamp.unix,
            timeline: {
                pct: {
                    x: 100
                }
            }
        };

        context.particle.data.timelineEvents = R.concat( context.particle.data.timelineEvents, [timelineEvent] );
    }

    return context;
};
