var shotsOnGoal = require('./rbb-shots-on-goal'),
    utils = require('./utils'),    
    R = require('ramda');

var assemblerType = 'rbb-shot-path';
exports.particleType = 'rbb-shot-path';


// config for template conversions from shots on goal template
// Pixel coords in the template for mapping opta values to template positions for the goal
var templateGoalZones = shotsOnGoal.templateGoalZones;
var templatePitchZones = shotsOnGoal.templatePitchZones;
var goalYScale =  shotsOnGoal.goalYScale;
var goalZScale =  shotsOnGoal.goalZScale;
var pitchXScale =  shotsOnGoal.pitchXScale;
var pitchYScale =  shotsOnGoal.pitchYScale;


function getAngle(startPt,endPt) {
    var angleRadians = Math.atan2(startPt.y - endPt.y, startPt.x - endPt.x);
    // angle in degrees
    return angleRadians * 180 / Math.PI;
}

function getDistance(startPt, endPt) {
    var x = endPt.x - startPt.x;
    var y = endPt.y - startPt.y;
    return  Math.sqrt( (x*x) + (y*y) );
}

function getLinePoint(angle, distance, pt) {
    return { 
        x: Math.round(Math.cos(angle * Math.PI / 180) * distance + pt.x),  
        y: Math.round(Math.sin(angle * Math.PI / 180) * distance + pt.y)
    };
}

function getControlPoint(startPt, endPt, qualifiers, isShadow) {
    var angle = getAngle( startPt, endPt );
    var dist = getDistance( startPt, endPt );
    var distanceMod = 0.5;

    var directionDif = 0;

    // don't curve shadows as much as shot paths
    var curveMod = 0.7;

    if( dist > 20 ) {
        if(!isShadow) {    
            // modify the angle based on the direction of line to give a little curve to the line
            directionDif = (angle >= 90) ? 10 : -10; 
            curveMod = 1.0;       
        }

        var qHas = R.has( R.__, qualifiers );
        // modify the distance away from the end point for the control point based on event qualifiers
        if( qHas('swerve-left') ) {
            angle -= (30 * curveMod);
            distanceMod = 0.35;
        } else if( qHas('swerve-right', qualifiers) ) {
            angle += (30 * curveMod);
            distanceMod = 0.35;
        }
    } 
    
    return getLinePoint( angle + directionDif, dist * distanceMod, endPt );
}

function getCalculatedPoints( startPt, endPt, event ) {
    var qs = event.qualifiers;

    var shadowStartPt = R.clone( startPt );
    var endPtShadow = { x: endPt.x, y: ( R.has( 'goal-mouth-y-co-ordinate', qs ) ) ? 188 : endPt.y + 10 };
    var cpShadow = getControlPoint( shadowStartPt, endPtShadow, qs, true );

    var adjustedStartPt = R.clone( startPt );
    if( R.has('volley', qs) ) {
        adjustedStartPt.y -= 30;
    } else if( R.has('head', qs) ) {
        adjustedStartPt.y -= 70;
    }


    // control point (need first as affects the end point for drawing arrow)
    var cp = getControlPoint( adjustedStartPt, endPt, qs, false );
    var distance = ( getDistance( adjustedStartPt, endPt ) <= 15 ) ? 0 : 15;

    return {
        "shot-path": {
            start: adjustedStartPt,
            cp: cp,
            end: getEndPoint( distance, endPt, cp )
        },
        "shadow-path": {
            start: shadowStartPt,
            cp: cpShadow,
            end: endPtShadow
        }       
    };
}

function getEndPoint(distance, endPt, controlPt) {
    var angle = getAngle(controlPt,endPt);
    return { 
        x: Math.round(Math.cos(angle * Math.PI / 180) * distance + endPt.x),  
        y: Math.round(Math.sin(angle * Math.PI / 180) * distance + endPt.y)
    };
}

function otherSide(side) {
    return (side === 'home') ? 'away' : ( ( side === 'away' ) ? 'home' : null );
}


var normalizeShotType = R.compose( R.join('-'), R.split(' '), R.toLower, R.path(['type', 'name']) );

exports.initial = function (config) {
    var findTeam = utils.chooseTeamName (R.values(config.teams));

    return {
        particle: {
            title: " ",
            type: exports.particleType,
            assemblerType: assemblerType,
            version: config.version,
            uniqueId: [ assemblerType, 0 ].join ( '-' ),
            data: {
                side: null,
                events: [],
                lastEvents: [],
                isHome: config.side === 'home',
                homeTeamName: findTeam('home'),
                awayTeamName: findTeam('away'),
                eventType: null,
                eventPlayer: null,
                goalKeeper: null,
                home: {
                    score: 0,
                    penalties: 0
                },
                away: {
                    score: 0,
                    penalties: 0
                }
            }
        },
        teams: config.teams,
        players: {},
        events: [],
        goalKeepers: {
            home: null,
            away: null
        }
    };
};

exports.reducer = function (context, event) {    
    context.particle.data.eventType = event.type && event.type.name;
    context.particle.data.eventPlayer = utils.playerName(context, event);

    utils.extendTeamData(context, event);
    utils.extendPlayerData(context, event);


    var hasQualifier = R.has( R.__, event.qualifiers );

    // clear data (only holding shot as it exists at that point in time)
    context.particle.title = ' ';
    context.particle.data.lastEvents = [];

    var side = event.team && context.teams[event.team.id].side.toLowerCase();

    if ( event.type.name === 'Team set up' ) {
        var playerPositions = R.zip ( event.qualifiers.involved.value.split ( ', ' ), event.qualifiers['player-position'].value.split ( ', ' ) );
        var goalKeeperId = parseInt ( R.find ( function ( positionPair ) {
            return positionPair[1] === '1';
        }, playerPositions )[0], 10 );
        context.goalKeepers[side] = utils.playerNameById ( context, goalKeeperId );
    }

    if ( event.type.name === 'Player on' ) {
        if ( R.path ( [ 'qualifiers', 'player-position', 'value' ], event ) === 'Goalkeeper' ) {
            context.goalKeepers[side] = utils.playerName ( context, event );
        }
    }
   
    context.particle.data.goalKeeper = context.goalKeepers[otherSide(side)];

    var didReachGoalLine = hasQualifier('goal-mouth-y-co-ordinate');
    if ( R.not(utils.isShot(event)) || R.not( didReachGoalLine )  ) {
        return context;
    }

    if ( event.type.name === 'Goal' ) {
        if ( event.period_id === 5 ) {
            if ( R.has ( 'own-goal', event.qualifiers ) ) {
                context.particle.data[otherSide(side)].penalties++;
            } else {
                context.particle.data[side].penalties++;
            }
        } else {
            if ( R.has ( 'own-goal', event.qualifiers ) ) {
                context.particle.data[otherSide(side)].score++;
            } else {
                context.particle.data[side].score++;
            }
        }
    }

    var findTeam = utils.chooseTeamName (R.values(context.teams));

    context.particle.title = R.join(' ', [
        event.qualifiers['own-goal'] ? 'Own goal:' : ( event.type.name + ':' ),
        findTeam( side ),
        '(',
        utils.getFormattedTime(event),
        ')'
    ]);


    var isBlocked = hasQualifier('blocked');

    var player = context.players[event.player.id];
    var jersey =  player ? player.jersey_number : '';
    var direction = context.teams[event.team.id].direction;
      

    var endCoords = {};
    var goalMouthCoords = {};


    // migrate to new namespace
    var template = {
        'goal-mouth-coords': {},
        'blocked-coords': {},
        'shot-path': {
            start: {},
            end: {},
            cp: {}
        },
        'shadow-path': {
            start: {},
            end: {},
            cp: {}
        }       
    };


    endCoords.x = 100;
    endCoords.y = event.qualifiers['goal-mouth-y-co-ordinate'].value;

    // remap the coordinates to one of 3 scales depending on the y value
    template['goal-mouth-coords'].x = goalYScale( event.qualifiers['goal-mouth-y-co-ordinate'].value, templateGoalZones );
    template['goal-mouth-coords'].y = goalZScale( event.qualifiers['goal-mouth-z-co-ordinate'].value, templateGoalZones );
    

    var eX = ( hasQualifier('own-goal') ) ? 100 - event.x : event.x;
    var eY = ( hasQualifier('own-goal') ) ? 100 - event.y : event.y;

    var points = getCalculatedPoints(
        { x: pitchYScale( eY, templatePitchZones ), y: pitchXScale( eX, templatePitchZones ) },
        { x: goalYScale( event.qualifiers['goal-mouth-y-co-ordinate'].value, templateGoalZones ), y: goalZScale( event.qualifiers['goal-mouth-z-co-ordinate'].value, templateGoalZones ) },
        event
    );

    template['shot-path'] = points['shot-path'];
    template['shadow-path'] = points['shadow-path'];
    

    var blockedCoords = {};
    if( isBlocked ) {
        blockedCoords.x = utils.translateX(direction, event.qualifiers['blocked-x-co-ordinate'].value);
        blockedCoords.y = utils.translateY(direction, event.qualifiers['blocked-y-co-ordinate'].value);            
    }

    var qualifiers = utils.getEventQualifiersAsBooleans(event);
    var shotType = normalizeShotType( event );

    var shotObj = {
        id: event.id,
        type: 'shot ' + shotType,
        side: side,
        start: {
            x: utils.translateX(direction, event.x),
            y: utils.translateY(direction, event.y)
        },
        end: {
            x: utils.translateX(direction, endCoords.x),
            y: utils.translateY(direction, endCoords.y)
        },
        mid: [blockedCoords],
        "is-goal": shotType === 'goal',
        "did-reach-goal-line": didReachGoalLine,
        "is-blocked": isBlocked,
        "is-penalties": event.period_id === 5,
        "is-on-target": utils.isShotOnTarget(event),
        template: template,
        qualifiers: qualifiers,
        original: event

    };

    context.particle.data.lastEvents = [ shotObj ];
    context.particle.uniqueId = [ assemblerType, event.id ].join ( '-' );

    return context;
};
