var utils = require('./utils');
var R = require('ramda');

var sortByGoal = R.sortBy( R.prop('is-goal') );

var assemblerType = 'rbb-shots-on-goal';
exports.assemblerType = assemblerType;
exports.particleType = 'rbb-shots-on-goal';

// mapping an input value from one range to another
function mapNumberRange ( num, in_min , in_max , out_min , out_max ) {
    return ( num - in_min ) * ( out_max - out_min ) / ( in_max - in_min ) + out_min;
}

// Pixel coords in the template for mapping opta values to template positions for the goal
var templateGoalZones = {
    left: { l: 0, r: 80  },
    closeLeft: { l: 80, r: 130 },
    postLeft: { l: 138, r: 142 },
    goal: { l: 150, r: 352, t: 90, b: 180 },
    postRight: { l: 362, r: 366 },
    closeRight: { l: 372, r: 410 },
    right: { l: 420, r: 500 },
    post: { t: 86, b: 88 },
    closeHigh: { t: 40, b: 78 },
    high: { t: 0, b: 40 }
};
exports.templateGoalZones = templateGoalZones;

// Pixel coords in the template for mapping opta values to template positions for the pitch
var templatePitchZones = {
    bounds: { l: 0, r: 500, t: 0, b: 350 },
    box6: { l: 80, r: 420, t: 188, b: 230 },
    box18: { l: -150, r: 650, t: 230, b: 280 },
    outside18: { l: -250, r: 700, t: 280, b: 320 },
    far: { l: -250, r: 700, t: 350, b: 350 }
};
exports.templatePitchZones = templatePitchZones;

// TODO Link to documentation about template
// There are 3 different scales we're operating on to try to show all data, without shrinking down goal area

// Y as relates to goal mouth is visually displayed as left/right
// the numbers relate to the opta mappings of goal (their Z axis): http://www.optasports.com/media/720876/appendix-12_560x350.jpg
// mapping it to template values (zone) to keep it in the correct zone
var goalYScale = function( value, zone ) {
    // raw value is 0-100 right to left, converting to left to right
    var invertedValue = 100 - value;

    if( invertedValue >= 45.2 && invertedValue <= 54.8 ) {
        // within goal area
        return mapNumberRange( invertedValue, 45.2, 54.8, zone.goal.l, zone.goal.r ); 

    } else  if( invertedValue >= 0 && invertedValue <= 40.7 ) {
        // far miss left
        return mapNumberRange( invertedValue, 0, 40.7, zone.left.l, zone.left.r );

    } else  if( invertedValue >= 40.7 && invertedValue <= 44.2 ) {
        // close miss left
        return mapNumberRange( invertedValue, 40.7, 44.2, zone.closeLeft.l, zone.closeLeft.r );

    } else  if( invertedValue > 44.2 && invertedValue < 45.2 ) {
        // left post
        return mapNumberRange( invertedValue, 44.2, 45.2, zone.postLeft.l, zone.postLeft.r );
        
    } else  if( invertedValue > 54.8 && invertedValue < 55.8 ) {
        // right post
        return mapNumberRange( invertedValue, 54.8, 55.8, zone.postRight.l, zone.postRight.r );

    } else  if( invertedValue >= 55.8 && invertedValue <= 59.3 ) {
        // close miss right
        return mapNumberRange( invertedValue, 55.8, 59.3, zone.closeRight.l, zone.closeRight.r );

    } else {
        // far miss right (59.3 to 100)
        return mapNumberRange( invertedValue, 59.3, 100, zone.right.l, zone.right.r );
    }
   
};
exports.goalYScale = goalYScale;

// Z as relates to goal mouth is visually displayed as up/down
// the numbers relate to the opta mappings of goal (their Z axis): http://www.optasports.com/media/720876/appendix-12_560x350.jpg
// mapping it to template values (zone) to keep it in the correct zone
var goalZScale = function( value, zone ) {
    if( value >= 60) {
        // far outside
        return mapNumberRange( 100 - value, 0, 40, zone.high.t, zone.high.b );
    } else  if( value > 42 ) {
        // closer miss
        return mapNumberRange( 100 - value, 40, 60, zone.closeHigh.t, zone.closeHigh.b );
    } else  if( value >= 38 ) {
        // post
        return mapNumberRange( 100 - value, 58, 62, zone.post.t, zone.post.b );
    } else {
        // within height of goal
        return mapNumberRange( 100 - value, 62, 100, zone.goal.t, zone.goal.b );
    }
};
exports.goalZScale = goalZScale;

// named as X as it uses the opta X value (which means the length of the pitch). Visually it will to/fro the goal on the pitch
// the numbers relate to the opta mappings of pitch length (their x axis): http://www.optasports.com/media/720871/appendix-11_560x360.jpg
// mapping it to template values (zone) to keep it in the correct zone
var pitchXScale = function(value, zone) {
    if( value >= 94.2) {
        // 6 yard box
        return mapNumberRange( (100 - value), 0, (100 - 94.2), zone.box6.t, zone.box6.b);
    } else  if( value >= 83) {
        // 18 yard box
        return mapNumberRange( (94.2 - value), 0, (94.2 - 83), zone.box18.t, zone.box18.b);
    } else  if( value >= 80) {
        // outside the 18 yard box to roughly the D
        return mapNumberRange( (83 - value), 0, (83 - 80), zone.outside18.t, zone.outside18.b);
    } else {
        // all the long blocked shots, put them part-way off the bottom of the screen
        return zone.bounds.b;
    }
};
exports.pitchXScale = pitchXScale;


// named as Y as it uses the opta Y value (which means the width of the pitch). Visually it will be left to right on the pitch
// the numbers relate to opta mappings of pitch width (their y axis) http://www.optasports.com/media/720871/appendix-11_560x360.jpg
// mapping it to template values (zone) to keep it in the correct zone
var pitchYScale = function(value, zone) {
     // raw value is 0-100 right to left, converting to left to right
    var invertedValue = 100 - value;

    if( invertedValue >= 36.8 && invertedValue <= 63.2 ) {
        // within 6 yard box
        return mapNumberRange( invertedValue, 36.8, 63.2, zone.box6.l, zone.box6.r );

    } else  if( invertedValue >= 21.1 && invertedValue <= 78.9 ) {
        // within 18 yard box
        return utils.clamp( mapNumberRange( invertedValue, 21.1, 78.9, zone.box18.l, zone.box18.r ), zone.bounds.l, zone.bounds.r );

    } else {
        // far block
        return  utils.clamp( mapNumberRange( invertedValue, 0, 100, zone.far.l, zone.far.r), zone.bounds.l, zone.bounds.r );
    }
};
exports.pitchYScale = pitchYScale;


exports.initial = function (config) {

    var findTeam = utils.chooseTeamName (R.values(config.teams));

    return {
        particle: {
            title: R.join(' ', [ findTeam( 'home' ), 'vs', findTeam( 'away' ), 'shots on goal' ]),
            type: exports.particleType,
            assemblerType: assemblerType,
            version: config.version,
            data: {
        		home: {
        		    name: findTeam('home'),
        		    events: [],
                    penalties:[],
                    passes: 0,
                    passesNormalised: utils.normaliseNumber ( 0 ),
                    passesPerShot: 0,
                    'cumulative-shots': {
            			'goals': 0,
            			'match': 0,
                        'match-on-target': 0,
            			'match-off-target': 0            		
                    },
                    'cumulative-shots-normalised': {
            			'goals': 0,
            			'match': 0,
                        'match-on-target': 0,
            			'match-off-target': 0
                    }
        		},
        		away: {
        		    name: findTeam('away'),
        		    events: [],
                    penalties: [],
                    passes: 0,
                    passesNormalised: utils.normaliseNumber ( 0 ),
                    passesPerShot: 0,
                    'cumulative-shots': {
            			'goals': 0,
            			'match': 0,
                        'match-on-target': 0,
            			'match-off-target': 0
                    },
                    'cumulative-shots-normalised': {
            			'goals': 0,
            			'match': 0,
                        'match-on-target': 0,
            			'match-off-target': 0
                    }
        		},
        		// for title / text templates:
        		venue: config.venue,
        		homeScore: 0,
        		awayScore: 0,
        		// TODO: this is duplicated in
        		// context.particle.data.home.name and
        		// context.particle.data.away.name, if title and text
        		// templates can be updated to use those, this code
        		// can be removed:
        		teams: {
                    home: {
                        rbbName: findTeam('home')
                    },
                    away: {
                        rbbName: findTeam('away')
                    }
                },

            }
        },
        teams: config.teams,
        players: {},
        gameTime: null,
    	scores: {
    	    home: 0,
    	    away: 0
    	},
    	home: {
    	    events: []
    	},
    	away: {
    	    events: []
    	}
    };
};

function didReachGoalLine(event) {
    return event.qualifiers.hasOwnProperty('goal-mouth-y-co-ordinate');
}

function isBlocked(event) {
    return event.type.name === 'Attempt Saved' && event.qualifiers.hasOwnProperty('blocked') && !didReachGoalLine(event);
}

function endCoords(event) {

    if( R.has('blocked-x-co-ordinate', event.qualifiers) ) {
        return blockedCoords( event ); 
    }

	return {
	    x: 100,
	    y: event.qualifiers['goal-mouth-y-co-ordinate'].value
	};
    
}

function blockedCoords(event) {
    return {
    	x: event.qualifiers['blocked-x-co-ordinate'].value,
    	y: event.qualifiers['blocked-y-co-ordinate'].value
    };
}

function translateCoords(direction, obj) {
    return {
        x: utils.translateX(direction, obj.x),
        y: utils.translateY(direction, obj.y)
    };
}

function convertedCoords(event) {

    // This gets a little confusing as opta use 
    //  x to mean goal to goal direction
    //  y to mean sideline to sideline
    //  z to mean up and down
    //  
    // We're mapping everything to x/y template coordinates where 
    //  x is sideline to sideline
    //  y is up/down for goal mouth and towards the goal for blocked shots plotted on ground

    if (didReachGoalLine(event)) {

    	return {
    	    'goal-mouth-coords': {        
        		x: goalYScale(
        		    event.qualifiers['goal-mouth-y-co-ordinate'].value,
        		    templateGoalZones
        		),
        		y: goalZScale(
        		    event.qualifiers['goal-mouth-z-co-ordinate'].value,
        		    templateGoalZones
        		)
    	    },
    	    'blocked-coords': {}
    	};
    }

    return {
	   'goal-mouth-coords': {},
	   'blocked-coords': {
    	    x: pitchYScale(
        		event.qualifiers['blocked-y-co-ordinate'].value,
        		templatePitchZones
        	    ),
            y: pitchXScale(
        		event.qualifiers['blocked-x-co-ordinate'].value,
        		templatePitchZones
        	    )
        }
    };
}



var normalizeShotType = R.compose( R.join('-'), R.split(' '), R.toLower, R.path(['type', 'name']) );

/*
 Notes:

 "Goal Mouth" refers to the goal line, not just the goal area
 
 There are some shots that don't make it all the way to the goal line
 Some shots have a blocked location and a goal line location

 Appendix 12 on opta: http://www.optasports.com/praxis/documentation/football-feed-appendices/f24-appendices.aspx

 Goal mouth is the goal line from 0 to 100, going from right to left when looking at goal
 To work with templates, we need to have it from 0 to 100 left to right, but also scaled to match the scale of templates
 (see vars at top)
 */        

function makeShot(context, event) {
    var side = context.teams[event.team.id].side.toLowerCase();
    var player = context.players[event.player.id];
    var playerName = ( player && player.name ) ? player.name : '';
    var jersey =  player ? player.jersey_number : '';
    // normalize direction depending on side
    var direction = (side === 'home') ? 'Right to Left' : 'Left to Right'; 
    var otherDirection = (direction === 'Right to Left') ? 'Left to Right' : 'Right to Left';
    var qualifiers = utils.getEventQualifiersAsBooleans(event);
    var isOwnGoal = utils.isOwnGoal( event );

    return {
        id: event.id,
        eid: event.event_id,
        type: normalizeShotType( event ),
        side: side,
        player: {
            jersey: jersey,
            name: playerName,
            display: true
        },
        start: translateCoords( direction, event),
        end: translateCoords( (isOwnGoal) ? otherDirection : direction, endCoords(event)),
        formattedTime: utils.getFormattedTime ( event ),
        "is-goal": utils.isGoal(event),
        "is-own-goal": isOwnGoal,
        "did-reach-goal-line": didReachGoalLine(event),
        "is-blocked": isBlocked(event),
        "is-first-half": utils.isFirstHalf(event),
        "is-second-half": utils.isSecondHalf(event),
        "is-penalties": utils.isPenalties(event),
        "is-on-target": utils.isShotOnTarget(event),
        template: convertedCoords( event ),
        qualifiers: qualifiers
    };
}

exports.reducer = function (context, event) {    
    if ( event.type.name === 'Condition change' ) {
        return context;
    }

    if ( ! event.team ) {
        return context;
    }

    var side = context.teams[event.team.id].side.toLowerCase();

    utils.extendTeamData(context, event);
    utils.extendPlayerData(context, event);
        
    


    // for title / text templates
    context.particle.data.period = utils.periodName(event.period_id);
    context.particle.data.period_lowercase = context.particle.data.period.toLowerCase();

    // Goal counter for title / text templates
    if (event.type.name === 'Goal') {
        if (utils.isOwnGoal(event)) {
	       context.scores[utils.getOpposingSide(side)]++;
       } else {
	       context.scores[side]++;
	   }
    }

    context.particle.data.homeScore = context.scores.home;
    context.particle.data.awayScore = context.scores.away;
    
    if ( utils.isPass ( event ) ) {
        context.particle.data[side].passes++;
        context.particle.data[side].passesNormalised = utils.normaliseNumber(context.particle.data[side].passes);
        context.particle.data[side].passesPerShot = Math.round ( context.particle.data[side].passes * 10 / ( context.particle.data[side]['cumulative-shots']['match-on-target'] || 1 ) ) / 10;
    }
    
    if (utils.isShot(event)) {
        var graphicSide = context.teams[event.team.id].side.toLowerCase();
        if (utils.isOwnGoal(event)) {
            // display on opposite graphic if it's an own goal
            graphicSide = utils.getOpposingSide(graphicSide);
        }

        // create the shot object
        var shot = makeShot(context, event);

        // if penalty shootout, add it to penalties array but don't count toward other things
        if (utils.isPenalties(event) && !context.penalties) {
            context.particle.data[graphicSide].penalties.push( makeShot(context, event) ); 
            return context;
        }

        // add shot to events arrays
        
        context[graphicSide].events.push(shot);
        // sorted so goals are last in the array, drawn on top for template
            context.particle.data[graphicSide].events = sortByGoal(context[graphicSide].events);

        // cumulative shot data
        var shots = context.particle.data[graphicSide]['cumulative-shots'];
        var normShots = context.particle.data[graphicSide]['cumulative-shots-normalised'];
        // count number of goals
        shots.goals = context[graphicSide].events.filter(function(x) { return x['is-goal']; }).length;
        normShots.goals = utils.normaliseNumber ( shots.goals );
        // don't count the opposition's own goal as positive shot stats for this team
        if (!utils.isOwnGoal(event) && !utils.isPenalties(event)) {
            // set up the shot stat totals for rules engine to examine
            shots.match++;
            normShots.match = utils.normaliseNumber ( shots.match );
            if (utils.isShotOnTarget(event)) {
                shots['match-on-target']++;
                normShots['match-on-target'] = utils.normaliseNumber ( shots['match-on-target'] );
            } else {
                shots['match-off-target']++;
                normShots['match-off-target'] = utils.normaliseNumber ( shots['match-off-target'] );
            }
           
        }
       
    }


    return context;
};
