var R = require('ramda');    

exports.absoluteX = function (direction, x) {
    if (direction === 'Right to Left') {
        return (100 - Number(x));
    }
    return Number(x);
};

exports.absoluteY = function (direction, y) {
    if (direction === 'Left to Right') {
        return (100 - Number(y));
    }
    return Number(y);
};

exports.translateX = function (direction, x) {
    return exports.absoluteX(direction, x) / 100 * 524;
};

exports.translateY = function (direction, y) {
    return exports.absoluteY(direction, y) / 100 * 334;
};


exports.invert100AndMultiply = function(num, multiplier) {
    return (100 - num) * multiplier;
};

exports.clamp = function(num, min, max) {
  return Math.min(Math.max(num, min), max);
};

exports.getSemiFormattedTime = function ( event ) {
    var max = {
        '1': 45,
        '2': 90,
        '3': 105,
        '4': 120
    }[ event.period_id ] || Infinity;

    return ( event.min > max ? [ max, event.min - max ] : [ event.min ] ).join ( ' + ' );
};

exports.getFormattedTime = function ( event ) {
    return module.exports.getSemiFormattedTime ( R.assocPath ( [ 'min' ], event.sec === 0 ? event.min : ( event.min + 1 ), event ) ) + ' min';
};

exports.createGameTimeObject = function (event, curGametime) {
    if (event.period_id && event.min && event.sec) {
        return {
            period: event.period_id,
            min: event.min,
            sec: event.sec,
            formatted: exports.getFormattedTime(event)
        }; 
    } else {
        return curGametime;
    }
};

exports.getOpposingSide = function(side) {
    return (side.toLowerCase() === 'home') ? 'away' : 'home';
};

exports.oppositeDirection = function (direction) {
    return direction === 'Left to Right' ? 'Right to Left' : 'Left to Right';
};

exports.isGenericPassEvent = function (event) {
    var qs = event.qualifiers;
    return ( R.path ( [ 'type', 'name' ], event ) !== 'Deleted event' ) && qs.hasOwnProperty('pass-end-x') && qs.hasOwnProperty('pass-end-y');
};

exports.isPassChainStart = function (event, prev_team) {
    if (event.team.id !== prev_team) {
        return true;
    }
    return (event.qualifiers['penalty'] ||
            event.qualifiers['free-kick'] ||
            event.qualifiers['throw-in']);
};

exports.isPassChainEnd = function (event, prev_team) {
    return (event.type.name === 'Blocked Pass' ||
            event.type.name === 'Blocked Cross' ||
            event.type.name === 'Foul' ||
            event.type.name === 'Attempt Saved' ||
            event.type.name === 'Corner Awarded' ||
            event.type.name === 'Punch' ||
            event.type.name === 'Post' ||
            (event.type.name === 'Tackle' && event.outcome === 1));
};

exports.isPass = function(event) {
    return R.propEq ( 'type_id', 1 )( event ) && R.equals ( 0, R.length ( R.intersection ( [ 2, 107, 123 ], R.pluck ( 'qualifier_id', R.values ( R.prop ( 'qualifiers', event ) ) ) ) ) );
};

exports.isShot = function(event) {
    var n = event.type.name.toLowerCase();
    return (
        n === 'goal' || 
        n === 'miss' || 
        n === 'post' || 
        n === 'attempt saved'
    );
};

exports.isGoal = function (event) {
    var qualifiers = exports.getEventQualifiersAsBooleans(event);
    return exports.qualifyShotType(event, qualifiers) === 'goal';
};

exports.isOwnGoal = function (event) {
    return event.qualifiers.hasOwnProperty('own-goal');
};

exports.isFirstHalf = function (event) {
    return event.period_id === 1;
};

// note: second half includes extra time
exports.isSecondHalf = function (event) {
    return event.period_id > 1 && event.period_id < 5;
};

exports.isPenalties = function (event) {
    return event.period_id === 5;
};

exports.periodName = function (period_id) {
    return {
	'1': 'First half',
	'2': 'Second half',
	'3': 'First half (extra time)',
	'4': 'Second half (extra time)',
	'5': 'Penalty shoot-out',
	'14': 'Post game',
	'15': 'Pre-game',
	'16': 'Pre-match'
    }[period_id] || 'Unknown';
};

exports.isOfType = function(qualifiers, types) {
    var qs = R.intersection(types, R.keys(qualifiers));
    return R.not( R.isEmpty(qs) );
};

exports.isCornerOrFreeKick = function(event) {    
    return exports.isOfType( event.qualifiers, ['corner-taken', 'free-kick-taken'] );
};

exports.getEventQualifiersAsBooleans = function (event) {
    var qualifiers = {};
    for (var k in event.qualifiers) {
        var q = event.qualifiers[k];
        if (q.type.values === '') {
            qualifiers[k] = true;
        }
    }
    return qualifiers;
};

exports.qualifyShotType = function(event, qualifiers) {
    var n = event.type.name.toLowerCase();
    var t;
    if(n === 'goal') {
        t = 'goal';
    } else if(n === 'attempt saved') {
        t = 'saved';
    } else if(qualifiers['blocked-x-co-ordinate']) {
        t = 'blocked';
    } else if(n === 'post') {
        t = 'post';
    } else {
        t = 'shot';
    }
    return t;
};

exports.isShotOnTarget = function(event) {
    var n = event.type.name.toLowerCase();
    return (n === 'goal' || n === 'attempt saved' && R.has('goal-mouth-z-co-ordinate', event.qualifiers ) );
};

exports.isInAttackingHalf = function(event) {
    if( exports.isShot(event) ) {
        return ( event.qualifiers['own-goal'] ) ? false : true;
    } else if (exports.isGenericPassEvent(event) ){
        return (
            event.x >= 50 ||
            event.qualifiers['pass-end-x'].value >= 50
        );
    } else {
        return false;
    }

};


// current game period in format for rules
exports.setTimePeriodForRules = function(event, context) {
    var n = event.type.name.toLowerCase();
    return {
        "id":(n === 'end' && event.period_id === 1) ? 'half-time' : event.period_id
    };
};

exports.getOptaEventTimeStamp = function(event) {
    var ts = (event.timestamp && event.timestamp.unix) ? event.timestamp.unix * 1000 : null;
    return ts;
};

exports.extendTeamData = function (context, event) {
    var team = event.team && context.teams[event.team.id];
    if (event.type.name === 'Team set up') {
        team.involved = event.qualifiers['involved'].value.split(', ');
        team.jersey_numbers = event.qualifiers['jersey-number'].value.split(', ');
        team.player_positions = event.qualifiers['player-position'].value.split(', ');
    }
    else if (event.type.name === 'Start') {
        team.direction = event.qualifiers['direction-of-play'].value;
    }
};

exports.extendPlayerData = function (context, event) {
    if (!context.players) {
	context.players = {};
    }
    if (event.type.name === 'Team set up') {
	var team = context.teams[event.team.id];
        team.involved.forEach(function (id, i) {
            var player = R.find(R.propEq('id', Number(id)), team.players);
            var name = player.name;
            context.players[id] = {
                name:  name.known || name.first + ' ' + name.last,
                jersey_number: team.jersey_numbers[i] || '',
                position: team.player_positions[i],
                substitute: false,
                team: event.team.id
            };
        });
    }
    else if (event.type.name === 'Player on') {
	if (!context.players[event.player.id]) {
	    context.players[event.player.id] = {};
	}
        // double check we have the player's jersey number and position
	context.players[event.player.id].jersey_number = event.qualifiers['jersey-number'].value;
    context.players[event.player.id].position = event.qualifiers['player-position'].value;
	context.players[event.player.id].substitute = true;
    }
};

exports.chooseTeamName = R.curry ( function ( teams, side ) {
    var team = R.find ( function ( team ) {
        return R.toLower ( team.side ) === R.toLower ( side );
    }, teams );

    if ( team.short_club_name.length > team.name.length ) {
        return team.name;
    }

    if ( team.short_club_name === '' ) {
        return team.name;
    }

    if ( R.find ( function ( portion ) {
        return portion.length < 2;
    }, team.short_club_name.split ( ' ' ) ) ) {
        return team.name;
    }

    return team.short_club_name;
} );

exports.playerNameById = R.curry ( function ( context, playerId ) {
    var findPlayer = function ( playerId ) {
        return R.find ( function ( player ) {
            return player.id === playerId;
        }, R.concat ( R.values ( context.teams )[0].players, R.values ( context.teams )[1].players ) );
    };

    var player = findPlayer ( playerId );

    return ( player && player.name ) ? ( player.name.known || player.name.first + ' ' + player.name.last ) : '';
} );

exports.playerName = R.curry ( function ( context, event ) {
    return event.player ? exports.playerNameById ( context, event.player.id ) : '';
} );

// these events are simply excluded from the reduction
exports.excludeFromChain = function (event) {
    return R.contains(event.type.name, [
	'Aerial', // Aerial duel – 50/50 when the ball is in the air – outcome will represent whether the duel was won or lost
        'Deleted event', // Event has been deleted – the event will remain as it was originally with the same ID but will be resent with the type altered to 43.
        'Condition change', // Change in playing conditions
    ]);
};

// these events immediately break the chain
// the event itself will not appear in the chain
exports.breaksChain = function (event) {
    return R.contains(event.type.name, [
	'Start', // Start of a match period
	'Team set up', // Team line up information
	'Penalty faced', // Goalkeeper event; penalty by opposition
	'End', // End of a match period
    ]);
};

// these events always end a chain
// the event itself will appear at the end of the chain
exports.endsChain = function (event) {
    if (event.type.name === 'Corner Awarded') {
	// Ball goes out of play for a corner kick
	return event.outcome === 1;
    }
    if (event.type.name === 'Foul') {
	return event.outcome === 0;
    }
    // Shown each time the ball goes out of play for a throw-in or goal kick
    if (event.type.name === 'Out') {
	return event.outcome === 0;
    }
    return R.contains(event.type.name, [
	'Goal', // All goals
	'Claim', // Goalkeeper event; catching a crossed ball
	'Keeper pick-up' // Goalkeeper event; picks up the ball
    ]);
};

// these events might end a chain, if possesion swaps sides on the
// subsequent generic-pass event the event itself will appear at the
// end of the chain
exports.maybeEndsChain = function (event) {
    return R.contains(event.type.name, [
	'Tackle', // Tackle = dispossesses an opponent of the ball - Outcome 1 = win & retain possession or out of play, 0 = win tackle but not possession
	'Take On', // Attempted dribble past an opponent (excluding when qualifier 211 is present as this is 'overrun' and is not always a duel event)
	'Ball touch', // Used when a player makes a bad touch on the ball and loses possession. Outcome 1 – ball simply hit the player unintentionally. Outcome 0 – Player unsuccessfully controlled the ball.
	'Interception', // When a player intercepts any pass event between opposition players and prevents the ball reaching its target.
	'Attempt Saved', // Shot saved - this event is for the player who made the shot. Qualifier 82 can be added for blocked shot.
	'Save', // Goalkeeper event; saving a shot on goal. Can also be an outfield player event with qualifier 94 for blocked shot.
	'Blocked Pass', // Similar to interception but player already very close to ball
	'Dispossessed', // Player is successfully tackled and loses possession of the ball
	'Shield ball opp', // Defender uses his body to shield the ball from an opponent as it rolls out of play
	'Card', // Bookings - will have red, yellow or 2nd yellow qualifier plus a reason
	'Punch', // Goalkeeper event; ball is punched clear
	'Keeper Sweeper', // When keeper comes off his line and/or out of his box to clear the ball
	'Clearance', // Player under pressure hits ball clear of the defensive zone or/and out of play
	'Post',
	'Miss' // Any shot on goal which goes wide or over the goal
    ]);
};

exports.isTouchEvent = function ( event ) {
    /**
     *  All events with these type IDs: 
        1, 2, 3, 4 (outcome = 1), 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 41, 42, 50, 54, 61, 73, 74

        as per opta docs "Appendix 3 Useful Queries"
        http://www.optasports.com/praxis/documentation/football-feed-appendices/f24-appendices.aspx
     */

    return R.contains ( event.type_id, [ 1, 2, 3, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 41, 42, 50, 54, 61, 73, 74 ] ) || ( event.type_id === 4 && event.outcome === 1 );
};

exports.transformTouchThirds = R.curry ( function ( event, previousThirds ) {
    var initThird = function () {
        return {
            total: 0,
            percentage: 33
        };
    };

    if ( ! previousThirds ) {
        return {
            defence: initThird (),
            midfield: initThird (),
            attack: initThird ()
        };
    }

    var X = parseFloat ( event.x );

    var defenceTotal = previousThirds.defence.total + ( X < 33.33 ? 1 : 0 );
    var midfieldTotal = previousThirds.midfield.total + ( ( X >= 33.33 && X < 66.67 ) ? 1 : 0 );
    var attackTotal = previousThirds.attack.total + ( X >= 66.67 ? 1 : 0 );

    var defencePerc = Math.round ( ( defenceTotal * 100 ) / ( defenceTotal + midfieldTotal + attackTotal ) );
    var midfieldPerc = Math.round ( ( midfieldTotal * 100 ) / ( defenceTotal + midfieldTotal + attackTotal ) );
    var attackPerc = 100 - defencePerc - midfieldPerc;

    var thirds = {
        defence: {
            total: defenceTotal,
            percentage: defencePerc
        },
        midfield: {
            total: midfieldTotal,
            percentage: midfieldPerc
        },
        attack: {
            total: attackTotal,
            percentage: attackPerc
        }
    };

    return thirds;
} );

exports.normaliseNumber = function ( num ) {
    if ( num <= 10 ) {
        return [ 'zero', 'one', 'two', 'three', 'four', 'five', 'six', 'seven', 'eight', 'nine', 'ten' ][ num ];
    }

    return '' + num;
};
