var EventEmitter = require('events').EventEmitter;
var inherits = require('inherits');


function MemoryCache() {
    EventEmitter.call(this);
    this.setMaxListeners(100);
    this.hashes = [];
    this.data = {};
    this.pendingAnnotations = {};
};
inherits(MemoryCache, EventEmitter);

MemoryCache.prototype.destroy = function () {
    this.removeAllListeners();
};

MemoryCache.prototype.storeData = function (hash, data) {
    var self = this;
    this.data[hash] = data;
    var pending = this.pendingAnnotations[hash];
    if (pending) {
	pending.forEach(function (x) {
	    self.annotateStoredData(hash, x.key, x.data);
	});
	delete this.pendingAnnotations[hash];
    }
};

MemoryCache.prototype.annotateStoredData = function (hash, key, data) {
    var event = this.data[hash];
    if (!event.hasOwnProperty(key)) {
	event[key] = [];
    }
    event[key].push(data);
};

MemoryCache.prototype.annotate = function (hash, key, data) {
    if (this.data.hasOwnProperty(hash)) {
	// already got the event to annotate, update and reset to event's index
	this.annotateStoredData(hash, key, data);
	var index = this.hashes.indexOf(hash);
	this.emit('reset', index);
    }
    else {
	// not got the event yet, store in pending annotations until it arrives
	if (!this.pendingAnnotations[hash]) {
	    this.pendingAnnotations[hash] = [];
	}
	this.pendingAnnotations[hash].push({
	    key: key,
	    data: data
	});
    }
};

MemoryCache.prototype.append = function (hash, data) {
    if (!data) {
	throw new Error('MemoryCache.append missing data: ' + hash);
    }
    this.hashes.push(hash);
    this.storeData(hash, data);
    this.emit('update');
};

MemoryCache.prototype.insertBefore = function (before, hash, data) {
    if (!data) {
	throw new Error('MemoryCache.insertBefore missing data: ' + hash);
    }
    var index = this.hashes.indexOf(before);
    this.hashes.splice(index, 0, hash);
    this.storeData(hash, data);
    this.emit('reset', index);
};

MemoryCache.prototype.replace = function (old, hash) {
    if (!(hash in this.data)) {
	throw new Error('MemoryCache.replace: no data for hash ' + hash);
    }
    if (!(old in this.data)) {
	throw new Error('MemoryCache.replace: no data for hash ' + old);
    }
    var index = this.hashes.indexOf(old);
    this.hashes.splice(this.hashes.indexOf(hash), 1);
    this.hashes.splice(index, 1, hash);
    //delete this.data[old];
    this.emit('reset', index);
};

MemoryCache.prototype.remove = function (hash) {
    if (!(hash in this.data)) {
	throw new Error('MemoryCache.remove: no data for hash ' + hash);
    }
    var index = this.hashes.indexOf(hash);
    this.hashes.splice(index, 1);
    //delete this.data[hash];
    this.emit('reset', index);
};

MemoryCache.prototype.get = function (index) {
    var hash = this.hashes[index];
    if (hash in this.data) {
	return this.data[this.hashes[index]];
    }
    else {
	throw new Error('Event ' + index + ' (' + hash + ') not found in data cache');
    }
};

MemoryCache.prototype.size = function () {
    return this.hashes.length;
};

MemoryCache.prototype.range = function (begin, end) {
    if (end === null || end > this.hashes.length) {
        end = this.hashes.length;
    }
    if (begin < 0) {
	begin = 0;
    }
    if (end < begin) {
        throw new Error('MemoryCache.range: end index is before begin index');
    }
    var result = [];
    for(var i = begin; i < end; i++) {
        result.push({index: i, data: this.get(i)});
    }
    return result;
};

MemoryCache.prototype.toArray = function () {
    var self = this;
    return self.hashes.map(function (hash, i) {
	return {
	    index: i,
	    hash: hash,
	    data: self.data[hash]
	};
    });
};

module.exports = MemoryCache;
