var EventEmitter = require('events').EventEmitter;
var inherits = require('inherits');
var assemblers = require('../assemblers/index');
var RenderedParticle = require('./rendered-particle');


function ParticleAssembler(subscriber, publishSubscriber, cache, globalConfig, checkpoint) {
    EventEmitter.call(this);
    this.subscriber = subscriber;
    this.publishSubscriber = publishSubscriber || null;
    this.cache = cache;
    this.globalConfig = globalConfig || {};
    this.particles = [];
    this.checkpoint = checkpoint;

    var self = this;

    subscriber.on('event', function (event) {
        var data = event.event;
        switch (data.type) {
            case 'append':
                self.cache.append(data.hash, data.event);
                break;
            case 'insertBefore':
                self.cache.insertBefore(data.before, data.hash, data.event);
                break;
            case 'replace':
                self.cache.replace(data.old, data.hash);
                break;
            case 'remove':
                self.cache.remove(data.hash);
                break;
        }
    });

    publishSubscriber && publishSubscriber.on('event', function (event) {});

    cache.on('update', function () {
	self.particles.forEach(function (p) {
	    if (!p.paused) {
		p._update();
	    }
	});
	self.emit('update');
    });

    cache.on('reset', function (index) {
	self.particles.forEach(function (p) {
	    p.reset(index);
	});
	self.emit('update');
    });
};
inherits(ParticleAssembler, EventEmitter);


function merge(a, b) {
    var obj = {};
    for (var k in a) {
        obj[k] = a[k];
    }
    for (var k in b) {
        obj[k] = b[k];
    }
    return obj;
}

ParticleAssembler.prototype.destroy = function () {
    this.removeAllListeners();
    this.subscriber.disconnect();
    this.cache.destroy();
    this.particles.forEach(function (p) {
        p.destroy();
    });
};

ParticleAssembler.prototype.add = function (spec) {
    var particle = new RenderedParticle(this.cache, {
        subscriber: this.subscriber,
        assembler: getParticleModule(spec.type),
        config: merge(this.globalConfig, spec.config || {}),
        checkpoint: this.checkpoint,
        type: spec.type,
        start: 0,
        end: null
    });
    this.particles.push(particle);
    return particle;
};

function getParticleModule(name) {
    if (name in assemblers) {
        return assemblers[name];
    }
    throw new Error('No such assembler: ' + name);
}

module.exports = ParticleAssembler;
