var EventEmitter = require('events').EventEmitter;
var inherits = require('inherits');
var R = require ( 'ramda' );

var now = (typeof window !== 'undefined' &&
           window.performance.now.bind(window.performance)) ||
        require('performance-now');


function RenderedParticle(cache, options) {
    EventEmitter.call(this);
    this.setMaxListeners(100);

    this.cache = cache;
    this.type = options.type;
    this.assembler = options.assembler;
    this.subscriber = options.subscriber;
    this.config = options.config;
    this.start = options.start || 0;
    this.end = options.end || null;
    
    if (options.checkpoint === undefined) {
        this._checkpoint_size = 500;
    }
    else {
        this._checkpoint_size = options.checkpoint;
    }
    
    this.context = this.assembler.initial(this.config);
    this.checkpoints = [R.clone(this.context)];
    this.paused = false;
    
    this.reset(0);
};
inherits(RenderedParticle, EventEmitter);

RenderedParticle.prototype.destroy = function (index) {
    this.removeAllListeners();
};

RenderedParticle.prototype.reset = function (index) {
    if (index === undefined) {
	throw new Error('Reset must specify an index argument');
    }
    if (index > this.index) {
        this._update();
        return;
    }
    var last_good_index = index - 1;
    var ci = this.checkpointForIndex(last_good_index);
    // delete checkpoints after this one
    this.checkpoints = this.checkpoints.slice(0, ci + 1);
    // reset to last checkpoint
    this.context = R.clone(this.checkpoints[ci]);
    this.index = ci * this._checkpoint_size - 1;
    // rebuild from there
    this._update();
};

RenderedParticle.prototype._foldEvent = function (context, event) {
    //console.log(['_foldEvent', context, event]);
    this.index = event.index;
    var result = this.assembler.reducer(context, event.data);
    if ((this.index + 1) % this._checkpoint_size === 0) {
        this.checkpoints.push(R.clone(result));
    }
    return result;
};

// returns checkpoint number before given index
RenderedParticle.prototype.checkpointForIndex = function (index) {
    if (index < 0) {
	return 0;
    }
    if (this._checkpoint_size) {
	var size = Math.min(
	    index,
	    (this.checkpoints.length - 1) * this._checkpoint_size
	);
        return Math.floor(size / this._checkpoint_size);
    }
    return 0;
};

RenderedParticle.prototype._update = function () {
    //console.log(['RenderedParticle _update', this.type, this.config.name]);
    if (this.paused) {
        //console.log('paused, skip _update');
        return;
    }
    if (!this.subscriber.caughtUp()) {
        // still processing batch data
        //console.log('still processing batch data, skip _update');
        return;
    }
    if (this.index == this.cache.size() - 1) {
     	// already processed all events in cache
     	return;
    }
    var start_time = now();
    var events = this.cache.range(this.index + 1, this.end);
    this.context = events.reduce(this._foldEvent.bind(this), this.context);
    if (this.index == this.end) {
        this.pause();
    }
    this.emit('update', this.context.particle, (now() - start_time));
};

RenderedParticle.prototype.pause = function (index) {
    // console.log(['RenderedParticle pause', index, this.cache.get(index).type.name]);
    if (index !== undefined) {
	    index = Math.min(index, this.cache.size() - 1);
	    index = Math.max(index, 0);
        if (this.index !== index) {
            // unpause so we can update the particle
            this.paused = false;
            var curr_end = this.end;
            this.end = index + 1;
            if (this.index < curr_end) {
                this._update();
            }
            else {
                this.reset(index);
            }
            this.end = curr_end;
        }
    }
    else {
        // already paused
        return;
    }
    this.paused = true;
    this.emit('pause');
};

RenderedParticle.prototype.pauseAt = function ( path, value ) {
    var event = R.find(function(event){
        return R.path(path, event) == value;
    },this.cache.range(0, this.cache.size ()));

    if (event) {
        this.pause(event.index);
        return true;
    } else {
        return false;
    }
};

RenderedParticle.prototype.resume = function () {
    //console.log(['RenderedParticle resume', this]);
    if (this.paused) {
        this.paused = false;
        this._update();
        this.emit('resume');
    }
};

module.exports = RenderedParticle;
